import logging.config


def init_log(show_time=True):
    """
    Initialise logger

    Args:
        show_time (bool, optional): print time in log entries. Defaults to True.
    """

    if show_time:
        log_format = "[%(levelname)s:%(asctime)s] %(message)s"
    else:
        log_format = "%(levelname)s | %(message)s"

    MY_LOGGING_CONFIG = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default_formatter": {"format": log_format},
        },
        "handlers": {
            "stream_handler": {
                "class": "logging.StreamHandler",
                "formatter": "default_formatter",
            },
        },
        "loggers": {
            "logger": {
                "handlers": ["stream_handler"],
                "level": "INFO",
                "propagate": True,
            }
        },
    }

    logging.config.dictConfig(MY_LOGGING_CONFIG)



import click
from denovowest.simulation import simulation
from denovowest.annotations import annotate_cadd, annotate_custom, annotate_dbnsfp, annotate_vcf


@click.group()
def main():
    """CLI for denovowest"""
    pass


# Register subcommands
main.add_command(simulation.main, name="simulation")
main.add_command(annotate_cadd.annotate_cadd, name="annotate-cadd")
main.add_command(annotate_custom.cli, name="annotate-custom")
main.add_command(annotate_dbnsfp.annotate_dbnsfp, name="annotate-dbnsfp")
main.add_command(annotate_vcf.cli, name="annotate-vcf")



#!/usr/bin/env python

import sys
import pandas as pd
import click
import logging

from scipy.stats import chi2
from denovowest.utils.log import init_log


def set_constraints_loci(region, gene_rates, threshold, ratio):
    """
    Flag constrained loci

    Args:
        region (pd.Series): Informations about a constrained region
    """

    # Compute the probability that a random variable with chi-square distribution is greater than chisq_diff_null
    p_value = chi2.sf(region["chisq_diff_null"], df=1)

    # Leave constrained loci status to False if p_value or observed/expected ratio above thresholds
    if (p_value > threshold) or (region["obs_exp"] > ratio):
        return

    # Set constrained status to True to every loci in the region which has as missense consequence
    gene_rates.loc[
        (gene_rates.pos >= region.cds_start)
        & ((gene_rates.pos <= region.cds_end))
        & ((gene_rates.consequence == "missense")),
        "constrained",
    ] = True


@click.command()
@click.argument("rates")
@click.argument("geneconstraint")
@click.argument("regionconstraint")
@click.argument("outrates")
def annotate_constraint(rates, geneconstraint, regionconstraint, outrates, threshold=1e-3, ratio=0.4):
    """Annotate rates file with the constrained status for each missense mutation loci in constrained regions

    Args:
        rates (str): path to rates file (already annotated with consequence)
        geneconstraint (str): 18K set inc different tx
        regionconstraint (str): >3k genes single entry
        outrates (str) : output path to constrained annotated rates file
        threshold (float, optional): _description_. Defaults to 1e-3.
        ratio (float, optional): _description_. Defaults to 0.4.
    """

    init_log()
    logger = logging.getLogger("logger")

    # Read constraints files
    geneconstraint_df = pd.read_table(geneconstraint)
    regionconstraint_df = pd.read_table(regionconstraint)

    # Read rates file and initialize all positions as non constrained
    rates_df = pd.read_csv(rates, sep="\t")
    if rates_df.empty:
        logger.warning("Rates file is empty")
        rates_df["constrained"] = None
        rates_df.to_csv(outrates, sep="\t", index=False)
        sys.exit(0)
    rates_df["constrained"] = False

    # For each gene in rates file
    for id_gene, gene_rates in rates_df.groupby("gene_id"):
        # We do not consider the version of the gene here
        ensembl_id = id_gene.split(".")[0]
        ensembl_id = ensembl_id.replace("gene:", "")

        # If the gene is not found in the constraint file, print as a warning
        if ensembl_id not in set(geneconstraint_df["ensembl_gene_id"]):
            logger.warning(f"Could not find id {ensembl_id} in full constraint file")
        # If the gene is found in the constraint file
        else:
            # Restrict the constraint data frame to current gene
            cur_geneconstraint_df = geneconstraint_df[geneconstraint_df["ensembl_gene_id"] == ensembl_id]
            assert cur_geneconstraint_df.shape[0] == 1

            # If there is only one constained region in this gene
            if cur_geneconstraint_df.iloc[0]["n_regions"] == 1:
                # Compute the probability that a random variable with chi-square distribution is greater than overall_chisq
                genep_value = chi2.sf(cur_geneconstraint_df["overall_chisq"], df=1)

                # If both this probability and the ratio of observed vs expected mutations is under their respective thresholds
                # we set all missense mutation loci constraint status to true for the current gene
                obs_exp_ratio = (
                    cur_geneconstraint_df.iloc[0]["obs_missense"] / cur_geneconstraint_df.iloc[0]["exp_missense"]
                )
                if (genep_value <= threshold) and (obs_exp_ratio <= ratio):
                    gene_rates.loc[gene_rates.consequence == "missense", "constrained"] = True
            # Otherwise if there are several constained regions in the gene
            else:
                # Restrict regional dataframe to current transcript
                transcript_id = cur_geneconstraint_df.iloc[0]["ensembl_transcript_enst"]
                cur_regionconstraint_df = regionconstraint_df[
                    regionconstraint_df["ensembl_transcript_enst"] == transcript_id
                ]

                # Set constrained status for each region
                for idx, region in cur_regionconstraint_df.iterrows():
                    set_constraints_loci(region, gene_rates, threshold, ratio)

            # Modify the rates data frame with constraints status
            rates_df.loc[gene_rates.index, "constrained"] = gene_rates.constrained

        # Export rates file (updated with constrained status)
        rates_df.to_csv(outrates, sep="\t", index=False)


if __name__ == "__main__":
    annotate_constraint()



#!/usr/bin/env python
import pandas as pd
import numpy as np
import click
import json
import pysam
import subprocess
import logging

from denovowest.utils.log import init_log


def rates_stats(rates_df):
    """
    Compute statistics such as missing annotations for each gene

    Args:
        rates_df (pd.DataFrame): rates
    """

    rates_stats_dict = dict()

    for gene_id, generates_df in rates_df.groupby("gene_id"):
        rates_stats_dict[gene_id] = gene_stats(gene_id, generates_df)

    return rates_stats_dict


def gene_stats(gene_id, generates_df):
    """
    Compute statistics such as missing annotations for each molecular consequence
    and each annotation.

    Args:
        gene_id (str) : gene identifier
        generates_df (pd.DataFrame): rates for a given gene
    """

    gene_dict = dict()

    for consequence, consequence_df in generates_df.groupby("consequence"):
        gene_dict[consequence] = consequence_stats(consequence, consequence_df)

    return gene_dict


def consequence_stats(consequence, consequence_df):
    """
    Compute statistics such as missing annotations for each molecular consequence
    and each annotation.

    Args:
        consequence (str): molecular consequence
        consequence_df (pd.DataFrame)
    """

    consequence_dict = dict()

    nb_mutations = consequence_df.shape[0]
    nb_missing_prob = consequence_df["prob"].isna().sum()
    consequence_dict["nb_mutations"] = int(nb_mutations)
    consequence_dict["nb_missing_prob"] = int(nb_missing_prob)

    scores = list(consequence_df.select_dtypes(include=["float"]).columns)

    for score in scores:
        score_dict = dict()
        nb_missing_score = consequence_df[score].isna().sum()
        deciles = consequence_df[score].quantile([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])

        score_dict["nb_missing_score"] = int(nb_missing_score)
        if nb_missing_score != nb_mutations:
            score_dict["deciles"] = list(deciles)

        consequence_dict[score] = score_dict

    return consequence_dict


def handle_nas(rates_df):
    """
    Turns "." and empty cells into NaNs in order to convert scores columns to float

    Args:
        rates_df (pd.DataFrame): rates
    """

    # TODO : Better handle this, the rates file should probably has missing values rather than "."
    rates_df.replace({".": np.nan, "": np.nan}, inplace=True)
    for column in rates_df.columns:
        if column in ["pos", "prob"]:
            continue
        try:
            rates_df[column] = rates_df[column].astype(float)
        except ValueError:
            continue

    return rates_df


def warning_genes(rates_stats_dict):
    """
    Find genes with too many missing scores for each annotation
    and log them.

    Args:
        rates_stats_dict (dict): annotation statistics
    """

    THRESHOLD_PERCENTAGE_MISSING_ANNOTATIONS = 0.1

    logger = logging.getLogger("logger")
    for gene_id, gene_stats in rates_stats_dict.items():
        perc_missing_annotations = percentage_missing_annotations(gene_stats)

        for score, percentage in perc_missing_annotations.items():
            if percentage > THRESHOLD_PERCENTAGE_MISSING_ANNOTATIONS:
                logger.warning(f"{gene_id} | {score} | {round(percentage * 100,2)}% missing annotations")


def percentage_missing_annotations(gene_stats):
    """
    Compute the percentage of missing scores for non-synonymous mutations.

    Args:
        gene_stats (dict): annotation statistics at the gene level

    Returns:
        dict: percentage of missing scores per annotation
    """

    consequences_to_consider = ["missense", "stop_gained", "splice_acceptor", "splice_donor", "start_lost", "stop_lost"]

    nb_mutation = 0
    nb_missing_annotations = dict()
    for consequence, consequence_stats in gene_stats.items():
        if consequence in consequences_to_consider:
            nb_mutation += consequence_stats["nb_mutations"]
            for score, score_stats in consequence_stats.items():
                if not (isinstance(score_stats, dict)):
                    continue
                if score in nb_missing_annotations.keys():
                    nb_missing_annotations[score] += score_stats["nb_missing_score"]
                else:
                    nb_missing_annotations[score] = score_stats["nb_missing_score"]

    for score, count in nb_missing_annotations.items():
        nb_missing_annotations[score] = count / nb_mutation

    return nb_missing_annotations


@click.command()
@click.argument("rates")
@click.argument("output")
@click.option("--fullgenome", is_flag=True)
def main(rates, output, fullgenome):
    """
    This script provides informations on an annotated mutation rates file
    such as the percentage of functional variants missing a score for each
    annotation.

    It is mainly dedicated to be run in the nextflow pipeline on subsetted
    rates file but can also be used on a full genome indexed rates file by using
    the --fullgenome flag.

    Args:
        rates (str): mutation rates file
        output (str): JSON file storing per-gene statistics
        fullgenome (boolean) : to use with full genome indexed mutation rates file
    """

    init_log()
    logger = logging.getLogger("logger")

    # If using a full genome indexed rates file
    if fullgenome:
        rates_file = pysam.TabixFile(rates, encoding="utf-8")

        # Retrieve rates file header
        cmd = f"zcat {rates} | head -n 1"
        res = subprocess.run(cmd, shell=True, capture_output=True, text=True)
        rates_file_columns = res.stdout.strip().split("\t")

        # Load one chromosome at a time (for memory sake) and compute stats
        chromosomes = ["chr" + str(x) for x in list(range(1, 23))] + ["chrX"]
        rates_stats_dict = dict()
        for chromosome in chromosomes:
            try:
                rates_chrom_df = pd.DataFrame(
                    [line.split("\t") for line in rates_file.fetch(chromosome)], columns=rates_file_columns
                )
            except ValueError:
                logger.warning(f"# {chromosome} not found in index")
                continue

            rates_chrom_df = handle_nas(rates_chrom_df)
            rates_chrom_stats_dict = rates_stats(rates_chrom_df)

            rates_stats_dict.update(rates_chrom_stats_dict)

    # Otherwise if using a subsetted rates file (e.g. from nextflow pipeline)
    else:
        rates_df = pd.read_csv(
            rates,
            sep="\t",
            dtype={
                "gene_id": str,
                "chrom": str,
                "pos": int,
                "ref": str,
                "alt": str,
                "consequence": str,
                "prob": float,
            },
        )

        rates_df = handle_nas(rates_df)
        rates_stats_dict = rates_stats(rates_df)

    # Export stats in a JSON file
    with open(output, "w") as f:
        json.dump(rates_stats_dict, f, indent=4)

    # Log genes with too many missing annotations
    warning_genes(rates_stats_dict)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import pandas as pd
import click
import pysam
import sys
import logging

from itertools import groupby, count
from denovowest.utils.log import init_log


def load_cadd(cadd_file, chrom, start, end):
    """
    Access to specific region in CADD file.
    Args:
        cadd_file (pysam.TabixFile): CADD score file
        chrom (str): chromosome identifier
        start (int): beginning position of the region
        end (int): terminating position of the region
    """

    def parse(line):
        chrom, pos, ref, alt, raw, scaled = line.split("\t")

        return {
            "chrom": chrom,
            "pos": int(pos),
            "ref": ref,
            "alt": alt,
            "raw": float(raw),
            "score": float(scaled),
        }

    return pd.DataFrame([parse(x) for x in cadd_file.fetch(chrom, start, end)])


def as_range(region):
    """
    Returns genomic region boundaries

    Args:
        region (list): genomic region positions

    Returns:
        tuple: beginning and terminating position of the genomic region
    """
    l = list(region)
    return l[0], l[-1]


@click.command()
@click.argument("rates")
@click.argument("cadd")
@click.argument("output")
def annotate_cadd(rates, cadd, output):
    """Adds CADD score to rates file

    Args:
        rates (str): Path to rates file
        cadd (str): Path to CADD file
        output (str): Path to output file (merged dataframe)
    """

    init_log()
    logger = logging.getLogger("logger")

    # Load rates file
    rates_df = pd.read_table(rates, dtype={"chrom": str, "pos": int, "ref": str, "alt": str})
    if rates_df.empty:
        logger.warning("Rates file is empty")
        rates_df["raw"] = None
        rates_df["score"] = None
        rates_df.to_csv(output, sep="\t", index=False)
        sys.exit(0)

    # Depending on the gff, chromosome can be defined as "chrX" or just "X"
    if str(rates_df.iloc[0].chrom).startswith("chr"):
        add_chr = True
    else:
        add_chr = False

    # Load cadd file
    cadd_df = pysam.TabixFile(cadd)

    # For each gene
    list_merged_df = list()
    for gene_id, gene_rates_df in rates_df.groupby("gene_id"):
        chrom = str(gene_rates_df.chrom.values[0]).replace("chr", "")

        # Split each gene in contiguous block (i.e. exons) and load CADD scores
        # (memory and performance issue)
        list_block_df = list()
        for _, block in groupby(sorted(set(gene_rates_df["pos"])), key=lambda n, c=count(): n - next(c)):
            start, end = as_range(block)
            try:
                block_cadd_df = load_cadd(cadd_df, chrom, start - 1, end)
                list_block_df.append(block_cadd_df)
            except ValueError:
                continue

        # Merge rates with CADD score
        if list_block_df:
            gene_cadd_df = pd.concat(list_block_df)
            if add_chr:
                gene_cadd_df.chrom = "chr" + gene_cadd_df.chrom

            merged_gene_df = gene_rates_df.merge(gene_cadd_df, how="left", on=["chrom", "pos", "ref", "alt"])
        else:
            merged_gene_df = gene_rates_df

        list_merged_df.append(merged_gene_df)

    # Combine results for all genes
    merged_df = pd.concat(list_merged_df)
    merged_df.to_csv(output, sep="\t", index=False)


if __name__ == "__main__":
    merged_df = annotate_cadd()



#!/usr/bin/env python
import pandas as pd
import json
import click
import os


@click.command()
@click.argument("json_stats")
@click.argument("outdir")
def main(json_stats, outdir):
    """
    This script loads a per-gene statistics JSON file containing informations
    (such as the number of missing annotation for a given CEP on a given gene
    and for a given type of molecular consequence) and creates data frame that summarize
    these statistics.

    Args:
        json_stats (str): per-gene statistics JSON file
        outdir (str): output directory
    """

    # Load JSON file containing statistics per gene
    with open(json_stats) as f:
        stats = json.load(f)

    # Create the output directory if not already existing
    os.makedirs(outdir, exist_ok=True)

    # Retrieve the list of all molecular consequences inferred by bcftools csq and the CEP used
    # to assign scores to every mutation in the rates file
    consequences = list_consequences(stats)
    ceps = list_ceps(stats)

    # Calculate the coverage for each gene for each CEP for each molecular consequence
    for consequence in consequences:
        coverage_per_consequence_df = coverage_per_consequence(stats, consequence, ceps)
        coverage_per_consequence_df.to_csv(f"{outdir}/coverage_{consequence.replace('&', '-')}.tsv", sep="\t")


def list_consequences(stats):
    """
    Return the list of bcftoolscsq consequences assigned to every locus
    in the rates file.

    Args:
        stats (dict): Per-gene statistics
    """

    list_consequences = list()
    for gene_id, gene_dict in stats.items():
        list_consequences += list(gene_dict.keys())

    list_consequences = list(set(list_consequences))

    return list_consequences


def list_ceps(stats):
    """
    Returns the list of Computational Effect Predictors (CEP) used to
    assign a score to every locus in the rates file

    Args:
        stats (dict): Per-gene statistics
    """

    list_ceps = list()
    for gene_id, gene_dict in stats.items():
        list_ceps += list(gene_dict["missense"].keys())

    list_ceps = list(set(list_ceps) - {"nb_mutations", "nb_missing_prob", "prob"})

    return list_ceps


def coverage_per_consequence(stats, consequence, ceps):
    """
    For each CEP, retrieves the percentage of loci annotated in a given
    gene for a given consequence

    Args:
        stats (dict): Per-gene statistics
        consequence (str): Molecular consequence (e.g. missense)
        ceps (list): List of CEP (e.g. AlphaMissense_score)
    """

    consequence_dict = {
        gene_id: gene_dict[consequence] if consequence in gene_dict.keys() else {}
        for (gene_id, gene_dict) in stats.items()
    }

    return compute_percentage_annotated(consequence_dict, ceps)


def compute_percentage_annotated(consequence_dict, ceps):
    """
    For each CEP, retrieves the percentage of loci annotated in a given
    gene for a given consequence

    Args:
        consequence_dict (dict): statistics for a given gene and a given type of molecular consequence
        ceps (list): list of CEP used to score the variants
    """

    percentage_annotated = dict()
    for gene_id, gene_data in consequence_dict.items():

        # If the current gene do not have any variants of the current consequence type we skip it
        if not gene_data:
            continue

        # Otherwise we calculate the percentage of loci annotated with every CEP used to annotate the rates file
        nb_mutations = gene_data["nb_mutations"]
        percentage_annotated[gene_id] = dict()
        for cep in ceps:
            if cep in gene_data.keys():
                percentage_annotated[gene_id][cep] = round(gene_data[cep]["nb_missing_score"] * 100 / nb_mutations, 2)
            else:
                percentage_annotated[gene_id][cep] = "."

    return pd.DataFrame(percentage_annotated).transpose()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# python enrichment.py ../../input/test/new_format/dnm_min.tsv ../../input/test/new_format/all_rates_min.tsv ../../input/weights_ppv_2020_01_17.tab --nmales 10000 --nfemales 10000
import logging
import os
import time

import click
import pandas as pd
import numpy as np

from denovowest.utils.log import init_log
from denovowest.utils.params import CONSEQUENCES_MAPPING, CONSEQUENCES_SEVERITIES
from denovowest.simulation.scores import prepare_scores
from denovowest.simulation.probabilities import get_pvalue


def load_dnm_rates(dnm, rates, column):
    """
    Load DNM and rates files and limit the analysis to genes shared by both files.
    When parallelising DNW on HPC, the rates file is split per gene.

    Args:
        dnm (str): path to observed DNM
        rates (str): path to rates file
        column (str): column that stores variant scores
    """

    logger = logging.getLogger("logger")

    dnm_df = pd.read_csv(dnm, sep="\t", dtype={"chrom": str, "pos": int, column: float}, na_values=[".", "NA"])
    rates_df = pd.read_csv(rates, sep="\t", dtype={"chrom": str, "pos": int, column: float}, na_values=[".", "NA"])

    shared_genes = set(dnm_df.gene_id.unique()) & set(rates_df.gene_id.unique())
    logger.info(
        f"{len(shared_genes)}/{dnm_df.gene_id.nunique()} genes with at least one observed DNM are found in the rates file"
    )

    dnm_df = dnm_df.loc[dnm_df.gene_id.isin(shared_genes)]
    rates_df = rates_df.loc[rates_df.gene_id.isin(shared_genes)]

    return dnm_df, rates_df


def prepare_dnm_rates(
    dnm_df: pd.DataFrame, rates_df: pd.DataFrame, column: str, nmales: int, nfemales: int, impute_missing: bool
):
    """
    Prepare DNM and rates file for the simulation

    Args:
        dnm_df (pd.DataFrame): DNM dataframe
        rates_df (pd.DataFrame): mutation rates dataframe
        column (str): column that stores variant scores
        nmales (int): number of males in the cohort
        nfemales (int): number of females in the cohort
        impute_missing (bool): whether to impute the variant with missing scores or not
    """
    # Filter on variant consequence and calculate cohort based expected mutation rates
    dnm_df = prepare_dnm(dnm_df)
    rates_df = prepare_rates(rates_df, nmales, nfemales)

    # Prepare scores (indel scores, imputation missing scores...)
    ctx = click.get_current_context()
    dnm_df, rates_df = prepare_scores(dnm_df, rates_df, column, ctx.params["runtype"], impute_missing)

    return dnm_df, rates_df


def prepare_dnm(dnm_df: pd.DataFrame):
    """
    Filter DNM file to remove functional consequence not handled.
    Assign a higher level consequence to each DNM.

    Args:
        dnmfile (str): DNM file
    """

    dnm_df = filter_on_consequences(dnm_df, "dnm")
    dnm_df = assign_meta_consequences(dnm_df)

    return dnm_df


def filter_on_consequences(df: pd.DataFrame, mode: str):
    """
    Filter all variants with a consequence not found in CONSEQUENCES_MAPPING

    Args:
        df (pd.DataFrame): variant table (rates or dnm) having a consequence column
        mode(str) : dnm or rates
    """

    logger = logging.getLogger("logger")

    # Extract the worst consequence but also keep the original consequence called by bcftoolscsq
    if mode == "dnm":
        df["original_consequence"] = df.consequence
    df.consequence = [extract_worst_consequence(csq) if isinstance(csq, str) else csq for csq in list(df.consequence)]

    # Filter variants depending on run type : non-synonymous or missense test
    ctx = click.get_current_context()
    if ctx.params["runtype"] == "ns":
        filt = df.consequence.isin(CONSEQUENCES_MAPPING.keys())
    else:
        filt = df.consequence.isin(["missense", "start_lost", "stop_lost"])

    kept_df = df.loc[filt]

    logger.info(f"{mode.upper()} - Before consequence filtering : {df.shape[0]} records")

    discarded_df = df.loc[~filt]
    if not discarded_df.empty:
        logger.warning(f"{discarded_df.shape[0]}/{df.shape[0]} records were discarded")
    else:
        logger.info("All records have an acceptable consequence.")

    logger.info(f"{mode.upper()} - After consequence filtering : {kept_df.shape[0]} records")

    return kept_df


def extract_worst_consequence(csq):
    """
    Bcftoolscsq sometimes return multiple consequences
    separated by an "&". We extract the worst one.

    Args:
        csq (str): consequences separated by "&"
    """

    list_csq = csq.split("&")
    worst_csq_value = 1
    worst_csq = ""
    for cur_csq in list_csq:
        if CONSEQUENCES_SEVERITIES[cur_csq] > worst_csq_value:
            worst_csq_value = CONSEQUENCES_SEVERITIES[cur_csq]
            worst_csq = cur_csq

    return worst_csq


def assign_meta_consequences(df: pd.DataFrame):
    """
    Assign a higher level consequence to each variant

    Args:
        df (pd.DataFrame): variant table (rates or dnm) having a consequence column
    """

    logger = logging.getLogger("logger")

    df.loc[:, "consequence"] = df.consequence.replace(CONSEQUENCES_MAPPING)

    consequence_counts = dict(df.consequence.value_counts())
    for consequence, count in consequence_counts.items():
        logger.info(f"{consequence} : {count}")

    return df


def prepare_rates(rates_df: pd.DataFrame, nmales: int, nfemales: int):
    """
    Load mutation rates file.

    Args:
        ratesfile (str): path to mutation rates file
    """

    rates_df = filter_on_consequences(rates_df, "rates")
    rates_df = assign_meta_consequences(rates_df)

    rates_df = compute_expected_number_of_mutations(rates_df, nmales, nfemales)

    return rates_df


def compute_expected_number_of_mutations(rates_df: pd.DataFrame, nmales: int, nfemales: int):
    """
    Adjust the expected number of mutations to the number of individuals in the cohort.

    Args:
        rates_df (pd.DataFrame): mutation rates dataframe
        nmales (int): number of males in the cohort
        nfemales (int): number of females in the cohort
    """

    # Compute the expected number of mutations as the product of the mutation rate and the number of individuals
    autosomal_factor = 2 * (nmales + nfemales)
    rates_df.loc[:, "prob"] = rates_df.prob * autosomal_factor

    # Apply an extra correction for the X chromosome
    x_factor_correction = compute_x_factor_correction(nmales, nfemales)
    rates_df.loc[:, "prob"] = np.where(
        rates_df.chrom.isin(["X", "chrX"]), x_factor_correction * rates_df.prob, rates_df.prob
    )

    return rates_df


def compute_x_factor_correction(nmales: int, nfemales: int):
    """
    Expected number of mutations on the X chromosome need to be adjusted to the number of male and female individuals.
    Scaling factors are computed using the alpha from the most recent SFHS (Scottish Family Health Study) phased de novo data.
    Correct the non-PAR chrX genes for fewer transmissions and lower rate (depends on alpha)

    Args:
        nmales (int): number of males in the cohort
        nfemales (int): number of females in the cohort
    """

    autosomal_factor = 2 * (nmales + nfemales)

    female_transmit = nmales + nfemales
    male_transmit = nfemales

    alpha = 3.4
    male_k = 2 / (1 + (1 / alpha))
    female_k = 2 / (1 + alpha)

    x_factor = ((male_transmit * male_k) + (female_transmit * female_k)) / autosomal_factor

    return x_factor


def run_simulations(dnm_df: pd.DataFrame, rates_df: pd.DataFrame, nsim: int, pvalcap: float, score_column: str):
    """
    For each gene in the DNM file, run nsim simulations and test whether or not this gene is significantly enriched in predictive DNM.

    Args:
        dnm_df (pd.DataFrame): DNM dataframe
        rates_df (pd.DataFrame): rates dataframe that contains all possible SNV
        nsim (int): number of simulations to run
        pvalcap (float): stop simulations if cumulative p-value > pvalcap
        score_column (str) : CEP scores
    """

    logger = logging.getLogger("logger")

    genes = dnm_df.gene_id.unique()
    results = []
    cpt = 0
    for gene in genes:
        simulation_results = run_simulation(rates_df, dnm_df, gene, nsim, pvalcap, score_column)
        if simulation_results:
            results.append(simulation_results)

        cpt += 1
        if cpt % 10 == 0:
            logger.info(f"Processed {cpt}/{len(genes)} genes")

    return results


def run_simulation(rates_df, dnm_df, gene_id, nsim, pvalcap, score_column):
    """
    Run nsim simulations and test whether or not gene gene_id is significantly enriched in predictive DNM

    Args:
        rates_df (pd.DataFrame): rates dataframe that contains all possible SNV for the given gene
        dnm_df (pd.DataFrame): DNM dataframe that contains all observed DNM for the given gene
        gene_id (str) : gene identifier
        nsim (int): number of simulations to run
        pvalcap (float): stop simulations if cumulative p-value > pvalcap
        score_column (str) : CEP scores
    """

    logger = logging.getLogger("logger")

    if gene_id not in rates_df.gene_id.unique():
        logger.debug("could not find " + str(gene_id))
        return

    logger.info(f"Testing {gene_id}")
    start_time = time.time()

    # Subset rates file to current gene
    generates = rates_df.loc[rates_df.gene_id == gene_id]

    # Sum the scores of all observed DNM in the gene
    nb_observed_mutations = dnm_df[dnm_df.gene_id == gene_id].shape[0]
    obs_sum_scores = dnm_df[dnm_df.gene_id == gene_id][score_column].sum()

    # Run nsim simulations
    pval, info, exp_sum_scores = get_pvalue(
        generates, obs_sum_scores, nsim, pvalcap, nb_observed_mutations, score_column
    )

    # Store how long the simulation took for each gene
    end_time = time.time()
    wall_time = end_time - start_time
    info = f"{info}|{wall_time:.6f}"

    # Return the gene id, its expected and observed sum of scores, the p-value from the enrichment simulation test and some informations about the simulation
    return (gene_id, exp_sum_scores, obs_sum_scores, pval, info)


def export_results(results: list, outdir: str):
    """
    Write enrichment results

    Args:
        results (list): list of per-gene enrichment simulation results
        outdir (str): output directory
    """

    os.makedirs(outdir, exist_ok=True)

    df = pd.DataFrame.from_records(results, columns=["symbol", "expected", "observed", "p-value", "info"])
    df["nb_sim"] = [x.split("|")[0] for x in df["info"]]
    df["nb_mutations_stop"] = [x.split("|")[1] for x in df["info"]]
    df["sequential_simulation"] = [x.split("|")[2] for x in df["info"]]
    df["log"] = [x.split("|")[3] for x in df["info"]]
    df["wall_time"] = [x.split("|")[4] for x in df["info"]]

    df.drop("info", axis=1, inplace=True)

    df.to_csv(f"{outdir}/enrichment_results.tsv", sep="\t", index=False)


def log_configuration(conf):
    """

    Log the configuration the simulation script was run with

    Args:
        conf (dict): script arguments
    """

    logger = logging.getLogger("logger")
    logger.info("DNW simulation running with the following parameters :")
    logger.info("----------")
    for key, value in sorted(conf.items()):
        logger.info(f"{key} : {value}")
    logger.info("----------")


@click.command()
@click.argument("dnm")
@click.argument("rates")
@click.argument("column")
@click.option("--nmales", required=True, type=int, help="Number of males individual in your cohort")
@click.option("--nfemales", required=True, type=int, help="Number of females individual in your cohort")
@click.option("--pvalcap", default=0.01, type=float, help="Stop simulations if cumulative p-value > pvalcap")
@click.option("--nsim", type=int, help="Minimum number of simulations for each gene", default=10**7)
@click.option(
    "--runtype",
    help="Run type is either missense test (mis) or non-synonymous (ns)",
    type=click.Choice(["ns", "mis"]),
    default="ns",
    show_default=True,
)
@click.option("--outdir", default="./")
@click.option(
    "--impute-missing-scores",
    is_flag=True,
    help="Impute missing scores by taking the median score for similar variants (i.e. same consequence) in the gene",
)
@click.option(
    "--jobs",
    default=1,
    help="Number of cores to use during simulation step",
)
def main(dnm, rates, column, nmales, nfemales, pvalcap, nsim, runtype, outdir, impute_missing_scores, jobs):
    """
    DeNovoWEST is a simulation-based method that tests for DNM enrichment in each gene separately.
    It uses computational effect predictors (CEP) scores to reflect the pathogenicity probability for each variant.


    Args:
        dnm (str): DNM file
        rates (str): Per-generation mutation rates file
        column (str): Column to use for the simulation
        nmales (int): Number of males individual in your cohort
        nfemales (int): Number of females individual in your cohort
        pvalcap (float): Stop simulations if cumulative p-value > pvalcap
        nsim (int): Minimum number of simulations for each gene
        runtype (str): Run type is either missense test (mis) or non-synonymous (ns)
        outdir (str): Output directory
        impute_missing_scores (bool) : Impute missing scores by taking the median score for similar variants (i.e. same consequence) in the gene
    """
    init_log()
    log_configuration(click.get_current_context().params)

    # Load DNM and rates files
    dnm_df, rates_df = load_dnm_rates(dnm, rates, column)

    # Prepare DNM and rates file for simulation
    dnm_df, rates_df = prepare_dnm_rates(dnm_df, rates_df, column, nmales, nfemales, impute_missing_scores)

    # Run simulations
    results = run_simulations(dnm_df, rates_df, nsim, pvalcap, column)

    # Export results
    export_results(results, outdir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import logging
import os
import sys

import click
import gffutils
import pandas as pd
import pyfaidx
import pysam
import glob

from denovowest.utils.params import CDS_OFFSET, ROULETTE_SCALING_FACTOR, CARLSON_SCALING_FACTOR
from denovowest.utils.log import init_log
from denovowest.utils.io_helpers import load_conf, superseed_conf


def load_mutation_rate_model(mutation_rate_model_file):
    """Load mutation rate model

    Args:
        mutation_rate_model_file (str): Path to a mutation rate model file.

    Returns:
        pd.DataFrame : The mutation rate model as a pandas data frame.

    Examples:
        The mutation rate model should follow the above format.
        from	to	mu_snp
        AAA	ACA	1.6681119889870802e-09
        AAA	AGA	2.9280578256688802e-09
        AAA	ATA	1.03370814817543e-09
    """
    mutation_rate_model = pd.read_csv(mutation_rate_model_file, sep="\t")

    mutation_rate_model.index = mutation_rate_model["from"] + "_" + mutation_rate_model["to"]

    return mutation_rate_model


def load_gene_list(conf, gff_db, column=0):
    """Load the optionally user provided gene of interest list.

    Args:
        conf (str): configuration
        gff_db (gffutils database) : GFF utils database
        column (str) : column to use to extract the genes identifiers

    Returns:
         list: List of genes identifiers.

    Examples:
        The gene list can be a simple list or a TSV file like this :
        symbol	ENSG	HGNC_ID
        TARDBP	ENSG00000120948.19	HGNC:11571
    """

    logger = logging.getLogger("logger")

    if "GENE_LIST" in conf.keys():
        logger.info(f"Getting gene list from : {conf['GENE_LIST']}")

        gene_list_file = conf["GENE_LIST"]
        # Simple list
        if column == 0:
            gene_list = list(pd.read_csv(gene_list_file, sep="\t", header=None).iloc[:, 0].values)
        # TSV file
        else:
            gene_list = list(pd.read_csv(gene_list_file, sep="\t").loc[:, column].values)
    else:
        logger.info(f"Getting all genes in : {conf['GFF']}")
        gene_list = list()
        for gene in gff_db.all_features(featuretype="gene", order_by="start"):
            gene_list.append(gene.attributes["ID"][0])

    return gene_list


def load_gff(gff_file, gff_db_path):
    """Create the gff database used by gffutils.

    Args:
        gff_file (str): Path to a GFF or a gffutils database file.
        gff_db_path (str): Path to the newly created gff database.

    Returns:
        gffutils.db: GFF database.

    """

    logger = logging.getLogger("logger")

    # gffutils db input
    if gff_file.endswith(".db"):
        logger.info(f"Loading gffutils database : {gff_file}")
        gff_db = gffutils.FeatureDB(gff_file)
    # GFF input
    else:
        try:
            os.remove(gff_db_path)
            logger.info(f"Removed old gffutil database : {gff_db_path}")
        except OSError:
            pass

        logger.info(f"Creating gffutil database : {gff_db_path}")
        gff_db = gffutils.create_db(gff_file, gff_db_path, merge_strategy="create_unique")

    return gff_db


def get_alternates(seq, range_model):
    """
    For a given kmer, returns all alternative kmers with a different central nucleotide.

    Args:
        seq (str): kmer string
        range_model (int): length of sequence on each side of the central nucleotide

    Returns:
        list: list of alternates codons
    """
    # We generate all possible SNPs
    list_alternates = list()
    for nucleotide in ["A", "C", "T", "G"]:
        alternate_seq = seq[:range_model] + nucleotide + seq[range_model + 1 :]
        list_alternates.append(alternate_seq)

    # Remove the WT sequence from the list
    list_alternates = list(set(list_alternates) - set([seq]))

    return list_alternates


def calculate_rates_cds(gene, start, mutation_rate_model, seq, range_model):
    """Calculate the rate of each possiblle single nucleotide mutation in a given sequence according to a mutation rate model

    Args:
        gene (gffutils.feature.Feature): Gene object from gff db
        start (int): Genomic coordinate of the start of the sequence
        mutation_rate_model (pd.DataFrame): Mutation rate model
        seq (str): Sequence of interest
        range_model (int): Length of sequence on each side of the central nucleotide

    Returns:
        list: List of mutation rates
    """

    # Get reverse complement if gene on reverse strand
    if gene.strand == "-":
        rev_seq = pyfaidx.complement(seq[::-1])
        cur_seq = rev_seq
    else:
        cur_seq = seq

    # Get the length of the current sequence
    len_seq = len(seq)

    list_mutations = list()
    # For each nucleotide in the sequence
    for i in range(range_model, len(cur_seq) - range_model):
        # We get the kmer centered on the current nucleotide
        ref = str(cur_seq[i - range_model : i + range_model + 1])

        # We generate all three possible alternate kmers changing the central nucleotide
        list_alternates = get_alternates(ref, range_model)

        # For the three alternate kmers we compute the mutation rate
        for alt in list_alternates:
            try:
                mutation_rate = mutation_rate_model.loc[ref + "_" + alt, "mu_snp"]
            except KeyError as e:
                # Ambiguous nucleotides
                continue

            ref_nuc = ref[range_model]
            alt_nuc = alt[range_model]

            # We update the genomic coordinates differently depending on the strand
            if gene.strand == "-":
                ref_nuc = pyfaidx.complement(ref_nuc)
                alt_nuc = pyfaidx.complement(alt_nuc)
                pos = start + len_seq - 1 - i
            else:
                pos = start + i

            # We build a Serie for each possible mutation at each loci
            s = pd.Series(
                {
                    "gene_id": gene.id,
                    "chrom": gene.chrom,
                    "pos": pos,
                    "ref": ref_nuc,
                    "alt": alt_nuc,
                    "prob": mutation_rate,
                }
            )

            # And append it to a list that we will use to build a data frame
            list_mutations.append(s)

    return list_mutations


def get_sequence(fasta, chrom, start, end):
    """Returns sequence from a fasta file according to genomic coordinates

    Args:
        fasta (pyfaidx.Fasta): Genome sequence
        chrom (string): chromosome id
        start (int): sequence start position
        end (int): sequence end position

    Returns:
        str: DNA sequence
    """

    return fasta[chrom][start:end]


def calculate_rates_kmer(mutation_rate_model, fasta, gff_db, gene_list):
    """Assign mutation rates to every considered loci using kmer mutation rate model.
    https://www.nature.com/articles/ng.3050


    Args:
        mutation_rate_model (pd.DataFrame): Mutation rate for every possible kmer transition
        fasta (pyfaidx.Fasta): Genome sequence
        gff_db (gffutils.db): GFF database
        gene_list (list) : list of genes to consider

    Returns:
        pd.DataFrame: a mutation rate data frame
    """

    logger = logging.getLogger("logger")

    # Default mutation model is 3-mer, but other models (5-mers, 7-mers...) can be used, hence we get the length here
    length_model = len(mutation_rate_model.iloc[0]["from"])

    # Number of neighboring nucleotides to consider when computing rates
    range_model = length_model // 2

    nb_genes = len(gene_list)
    logger.info(f"Start computing rates for {nb_genes} genes")

    list_mutation_rates = list()
    cpt = 1
    # Loop through all CDS of interest to generate mutation rates
    for gene in gff_db.all_features(featuretype="gene"):
        # Skip gene if not in user provided gene list
        gene_id = gene.attributes["ID"][0]
        if gene_id not in gene_list:
            continue

        # Store mutation rates for the current gene
        list_mutation_rates_gene = list()
        # Store CDS boundaries to avoid calculating rates for same CDS in several transcripts
        list_cds_boundaries = list()
        for transcript in gff_db.children(gene, level=1):
            for cds in gff_db.children(transcript, featuretype="CDS", order_by="start"):
                # We add an offset to CDS region to consider loci such as splicing sites (default is 50),
                # the range model so that we get the neighboring nucleotides and we correct for python 0-index
                start = cds.start - CDS_OFFSET - range_model - 1
                end = cds.stop + CDS_OFFSET + range_model

                # If the CDS has already been covered by another transcript we skip it
                if (start, end) in list_cds_boundaries:
                    continue
                else:
                    list_cds_boundaries.append((start, end))

                # We extract the coding sequence
                cds_seq = get_sequence(fasta, gene.chrom, start, end)
                # cds_seq = fasta[gene.chrom][start:end]

                # We calculate the rates for the current CDS region
                list_mutation_rates_cds = calculate_rates_cds(
                    gene, start + 1, mutation_rate_model, cds_seq, range_model
                )
                list_mutation_rates_gene += list_mutation_rates_cds

        # For each gene we build a data frame and remove possible duplicated values
        if list_mutation_rates_gene:
            mutation_rates_gene_df = pd.DataFrame(list_mutation_rates_gene)
            mutation_rates_gene_df.drop_duplicates(inplace=True, keep="first")
            mutation_rates_gene_df.sort_values(by=["pos", "alt"], inplace=True)
            list_mutation_rates.append(mutation_rates_gene_df)
        else:
            logger.warning(f"No mutations found for gene {gene_id}")

        # Log progress
        if cpt % 100 == 0:
            logger.info(f"{cpt}/{nb_genes} done")

        cpt += 1

    if (cpt - 1) != nb_genes:
        logger.warning(f"Rates computed for {cpt - 1} genes when number of genes in list is {nb_genes} ")
    else:
        logger.info(f"Rates computed for {nb_genes} genes")

    # We assemble all genes data frame together
    if list_mutation_rates:
        mutation_rates_df = pd.concat(list_mutation_rates)
    else:
        mutation_rates_df = pd.DataFrame(columns=["gene_id", "chrom", "pos", "ref", "alt", "prob"])
        logger.warning(f"Rates dataframe is empty")

    return mutation_rates_df


def roulette_per_chrom_files(roulette_dir):
    """
    Build a dictionnary with chromosome as key and path to roulette vcf file as value

    Args:
        roulette_dir (str): path to directory containing roulette vcf files per chromosome

    Returns:
        dict: a dictionnary with chromosome as key and path to roulette vcf file as value
    """

    roulette_per_chrom_files_list = glob.glob(f"{roulette_dir}/*all.vcf.bgz")
    roulette_per_chrom_files = dict()
    for f in roulette_per_chrom_files_list:
        chrom = f.split("/")[-1].split("_")[0]
        roulette_per_chrom_files[chrom] = f

    return roulette_per_chrom_files


def calculate_rates_roulette(roulette_dir, gff_db, gene_list, model):
    """
    Use Roulette mutation rate model to generate a rates file
    https://github.com/vseplyarskiy/Roulette
    https://www.biorxiv.org/content/10.1101/2022.08.20.504670v1

    The roulette directory contains one VCF file per chromosome.
    Loci in each VCF are annotated with several mutation rates :
        - MR : roulette mutation rate
        - MR : carlson mutation rate

    Args:
        roulette_dir (string): path to directory containing roulette vcf files per chromosome
        gff_db (gffutils.db): GFF database
        gene_list (list) : list of genes to consider
        model (str) : carlson or roulette

    Returns:
        pd.DataFrame: a mutation rate data frame
    """
    logger = logging.getLogger("logger")

    # Select the mutation rate info field and the per generation mutation rate scaling factor depending on the model
    if model == "roulette":
        info_field = "MR"
        scaling_factor = ROULETTE_SCALING_FACTOR
    else:
        info_field = "MC"
        scaling_factor = CARLSON_SCALING_FACTOR

    nb_genes = len(gene_list)
    logger.info(f"Start computing rates for {nb_genes} genes")

    # Build a dictionnary with chromosome as key and path to roulette vcf file as value
    roulette_vcfs = roulette_per_chrom_files(roulette_dir)

    list_mutation_rates = list()
    cpt = 1
    # Loop through all CDS of interest to generate mutation rates
    for gene in gff_db.all_features(featuretype="gene"):
        # Skip gene if not in user provided gene list
        gene_id = gene.attributes["ID"][0]
        if gene_id not in gene_list:
            continue

        # Load roulette vcf file corresponding to the current gene
        chrom = gene.chrom.replace("chr", "")

        try:
            roulette_file = pysam.VariantFile(roulette_vcfs[chrom], index_filename=f"{roulette_vcfs[chrom]}.csi")
        except KeyError as e:
            # Roulette does not provide mutation rates for allosomes
            if chrom in ["X", "Y"]:
                continue
            else:
                logger.warning(f"Can't find any roulette file corresponding to chromosome {chrom} for gene {gene_id}")

        # Store mutation rates for the current gene
        list_mutation_rates_gene = list()
        # Store CDS boundaries to avoid calculating rates for same CDS in several transcripts
        list_cds_boundaries = list()
        for transcript in gff_db.children(gene, level=1):
            for cds in gff_db.children(transcript, featuretype="CDS", order_by="start"):
                # We add an offset to CDS region to consider loci such as splicing sites (default is 50),
                start = cds.start - CDS_OFFSET - 1
                end = cds.stop + CDS_OFFSET

                # If the CDS has already been covered by another transcript we skip it
                if (start, end) in list_cds_boundaries:
                    continue
                else:
                    list_cds_boundaries.append((start, end))

                # Extract mutation rates for the current CDS region and multiply by the scaling factor
                list_mutation_rates_cds = list()
                for rec in roulette_file.fetch(start=start, stop=end, region=chrom):
                    try:
                        s = pd.Series(
                            {
                                "gene_id": gene_id,
                                "chrom": chrom,
                                "pos": rec.pos,
                                "ref": rec.ref,
                                "alt": rec.alts[0],
                                "prob": float(rec.info[info_field]) * scaling_factor,
                            }
                        )
                        list_mutation_rates_cds.append(s)
                    except KeyError as e:
                        # No mutation rate info for this SNP
                        continue

                list_mutation_rates_gene += list_mutation_rates_cds

        # For each gene we build a data frame and remove possible duplicated values
        if list_mutation_rates_gene:
            mutation_rates_gene_df = pd.DataFrame(list_mutation_rates_gene)
            mutation_rates_gene_df.drop_duplicates(inplace=True, keep="first")
            mutation_rates_gene_df.sort_values(by=["pos", "alt"], inplace=True)
            list_mutation_rates.append(mutation_rates_gene_df)
        else:
            logger.warning(f"No mutations found for gene {gene_id}")

        # Log progress
        if cpt % 100 == 0:
            logger.info(f"{cpt}/{nb_genes} done")

        cpt += 1

    if (cpt - 1) != nb_genes:
        logger.warning(f"Rates computed for {cpt - 1} genes when number of genes in list is {nb_genes} ")
    else:
        logger.info(f"Rates computed for {nb_genes} genes")

    # We assemble all genes data frame together
    if list_mutation_rates:
        mutation_rates_df = pd.concat(list_mutation_rates)
    else:
        mutation_rates_df = pd.DataFrame(columns=["gene_id", "chrom", "pos", "ref", "alt", "prob"])
        logger.warning(f"Rates dataframe is empty")

    return mutation_rates_df


def validate_model(ctx, param, value):
    """
    Validatin function called at the program start to check model is correct.
    Model can be either kmer, carlson or roulette

    Args:
        ctx (click.Context):  context
        param (click.Parameter): parameter
        value (str): value provided for the model

    Raises:
        click.BadParameter: raised if user provided a value not in the allowed values

    Returns:
        str: model value
    """
    allowed_models = ["kmer", "carlson", "roulette"]
    if value not in allowed_models:
        raise click.BadParameter(f'Invalid model. Allowed values: {", ".join(allowed_models)}')
    return value


@click.command()
@click.option("--config")
@click.option("--gff")
@click.option("--fasta")
@click.option("--mutation_rate_model", help="Path to a k-mer mutation rate model or roulette directory")
@click.option("--gene_list")
@click.option("--outdir")
@click.option("--model", type=click.STRING, callback=validate_model)
def main(config, gff, fasta, mutation_rate_model, gene_list, outdir, model):
    """
    Generate per generation mutation rates according to a mutation rate model.
    Loci to be considered are provided as a GFF file or a gffutils database file, and a gene list.

    Args:
        conf_file (str): Use a YAML configuration file instead of command line arguments.
        gff (str): Annotations provided as GFF or a gffutils database file.
        fasta (str): Genome sequence in fasta format to retrieve kmer sequences (kmer only)
        mutation_rate_model (str): Path to a k-mer mutation rate model or roulette directory
        gene_list (str): File containing a list of genes of interest.
        outdir (str): Output directory.
        model (str): Mutation rate model to be used. Allowed values: kmer, carlson or roulette
    """

    # Initiate logger
    init_log()
    logger = logging.getLogger("logger")
    logger.info("Running {}".format(__file__.split("/")[-1]))

    # Load configuration file
    if config:
        try:
            conf = load_conf(config)
        except FileNotFoundError:
            logger.error(f"No such config file {config}")
            sys.exit(1)
    else:
        conf = dict()

    # Superseed configuration in config file with the configuration passed through command line arguments
    ctx = click.get_current_context()
    conf = superseed_conf(conf, ctx.params)

    # Log configuration
    logger.info(f"Parameters :")
    logger.info("----------")
    for key, value in conf.items():
        logger.info(f"{key} : {value}")
    logger.info("----------")

    # Create output directory
    os.makedirs(conf["OUTDIR"], exist_ok=True)

    # Load GFF file or GFF database used by gffutils
    gff_db = load_gff(conf["GFF"], f'{conf["OUTDIR"]}/gff.db')

    # Load gene list
    gene_list = load_gene_list(conf, gff_db)

    if model == "kmer":
        # Load mutation rate model
        mutation_rate_model = load_mutation_rate_model(conf["MUTATION_RATE_MODEL"])

        # Load fasta file
        fasta = pyfaidx.Fasta(conf["FASTA"])

        # Calculate mutation rates for every loci of interest
        mutation_rates_df = calculate_rates_kmer(mutation_rate_model, fasta, gff_db, gene_list)
    else:
        mutation_rates_df = calculate_rates_roulette(mutation_rate_model, gff_db, gene_list, model)

    # Export mutation rates file
    mutation_rates_df.to_csv("{}/{}".format(conf["OUTDIR"], "mutation_rates.tsv"), sep="\t", index=None)
    logger.info(f'Mutation rates file created here : {conf["OUTDIR"]}/mutation_rates.tsv')


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import click
import gffutils
import pandas as pd
import sys
import logging

from denovowest.utils.log import init_log
from denovowest.utils.params import CONSEQUENCES_SEVERITIES


@click.group()
def cli():
    pass


@cli.command()
@click.argument("rates")
@click.argument("fasta")
@click.argument("out_vcf")
def rates_to_vcf(rates, fasta, out_vcf):
    """Transforms a rate file (tabular format) in VCF format to use before calling bcftoolscsq

    Args:
        rates (file): Rate file
        fasta (file): FASTA file or fasta index file used to compute the rate file
        out_vcf (str): Destination of the output VCF file
    """

    # Retrieve contig length from FASTA file
    if fasta.endswith(".fai"):
        vcf_header_df = pd.read_csv(f"{fasta}", sep="\t", header=None)
    else:
        vcf_header_df = pd.read_csv(f"{fasta}.fai", sep="\t", header=None)

    # Write it in the VCF header format
    with open(out_vcf, "w") as f:
        f.write("##fileformat=VCFv4.2\n")
        for idx, row in vcf_header_df.iterrows():
            chrom = row[0]
            length = row[1]
            f.write(f"##contig=<ID={chrom},length={length}>\n")
        f.write('##INFO=<ID=GENE,Number=.,Type=String,Description="Gene identifier">\n')
        f.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n")

    # Extract information from rates file and transform it in VCF format
    rates_df = pd.read_csv(rates, sep="\t")

    # Filter duplicate variants
    duplicates = rates_df[["gene_id", "chrom", "pos", "ref", "alt"]].duplicated()
    rates_df = rates_df[~duplicates]

    # We need to sort the rates file otherwise bcftoolscsq fails
    rates_df = rates_df.sort_values(["chrom", "pos"])

    # Write VCF (including gene id in INFO field to distinguish between overlapping genes)
    with open(out_vcf, "a") as f:
        for idx, row in rates_df.iterrows():
            f.write(f"{row.chrom}\t{row.pos}\t.\t{row.ref}\t{row.alt}\t.\t.\tGENE={row.gene_id}\n")


def extract_consequence(x, gff_db):
    # Get gene name(s)
    gene_id = x.split(";")[0].replace("GENE=", "")

    # TODO : Handle the case where this field is missing or has a different name, find a clever way to support all gff
    try:
        gene_names = gff_db[gene_id].attributes["gene_name"]
    except KeyError:
        gene_names = gff_db[gene_id].attributes["Name"]

    # gene_transcripts_id = list()
    # for transcript in gff_db.children(gff_db[gene_id], featuretype="transcript"):
    #     gene_transcripts_id.append(transcript.id)

    # Look at all the consequences returned by bcftoolscsq
    try:
        list_csq = x.split(";")[1].replace("BCSQ=", "").split(",")

        # If there are multiple consequences, we take the worst one that
        # corresponds to our gene of interest
        # (we have multiple consequences when using multiple transcripts gff or even when using single transcripts like MANE, within splice regions)
        worst_csq_value = 1
        worst_csq = ""
        for cur_csq in list_csq:
            if cur_csq.split("|")[1] in gene_names:
                csqs = cur_csq.split("|")[0]
                for csq in csqs.split("&"):
                    if CONSEQUENCES_SEVERITIES[csq] > worst_csq_value:
                        worst_csq_value = CONSEQUENCES_SEVERITIES[csq]
                        worst_csq = csqs

        if worst_csq:
            csq = worst_csq
        else:
            csq = ""

    except IndexError:
        # bcftoolscsq return sometimes empty consequences
        csq = ""

    return csq


@cli.command()
@click.argument("vcf")
@click.argument("rates")
@click.argument("gff_db")
@click.argument("out_rates")
def vcf_to_rates(vcf, rates, gff_db, out_rates):
    """Merge the VCF file annotated with bcftoolscsq with a rate file (tabular)
    Args:
        vcf (str): Path of the VCF file
        rates (str): Path to the rates file
        gff_db (str) : Path to gffutils database
        out_rates (str): Path of the output rates file
    """

    init_log()
    logger = logging.getLogger("logger")

    # Read rates
    rates_df = pd.read_csv(rates, sep="\t", dtype={"chrom": str})

    # Read VCF
    try:
        if vcf.endswith("gz"):
            vcf_df = pd.read_csv(vcf, sep="\t", comment="#", header=None, compression="gzip")
        else:
            vcf_df = pd.read_csv(vcf, sep="\t", comment="#", header=None)
        vcf_df.columns = "#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO".split("\t")
        vcf_df["#CHROM"] = vcf_df["#CHROM"].astype(str)

    except pd.errors.EmptyDataError:
        if rates_df.empty:
            logger.warning("Rates file is empty")
            rates_df["consequence"] = None
            rates_df.to_csv(out_rates, sep="\t", index=False)
            sys.exit(0)
        else:
            logger.error("Bcftools csq file is empty")
            sys.exit(1)

    # Load gffutils database
    gff_db = gffutils.FeatureDB(gff_db)

    # Extract BCFtools consequence in a separate column
    vcf_df["consequence"] = [extract_consequence(x, gff_db) for x in vcf_df.INFO]

    # Extract gene id to distinguish between overlapping genes
    vcf_df["gene_id"] = [x.split(";")[0].replace("GENE=", "") for x in vcf_df.INFO]

    # Merge files
    out_rates_df = rates_df.merge(
        vcf_df[["#CHROM", "POS", "REF", "ALT", "consequence", "gene_id"]],
        left_on=["chrom", "pos", "ref", "alt", "gene_id"],
        right_on=["#CHROM", "POS", "REF", "ALT", "gene_id"],
    )
    out_rates_df.drop(["#CHROM", "POS", "REF", "ALT"], axis=1, inplace=True)

    assert out_rates_df.shape[0] == rates_df.shape[0]

    # Export
    out_rates_df.to_csv(out_rates, sep="\t", index=False)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
import click
import pysam
import pandas as pd
import logging

from itertools import groupby, count

from denovowest.utils.log import init_log
from denovowest.utils.io_helpers import load_gff, read_columns_from_file


def is_chr_prefixed(df_path):
    """
    Check whether the chromosomes are prefixed with "chr" in the input tables

    Args:
        df_path (str): path to a variant or annotation file

    """

    df = pd.read_csv(df_path, sep="\t", comment="#", nrows=10000)

    # We first check for a chrom column
    for key in ["chrom", "chr", "#chr", "#chrom"]:
        if key in df.columns:
            is_chr_prefixed = str(df[key].iloc[0]).startswith("chr")
            break

    # If there is no chrom column we try with the first one
    # TODO : not optimal
    if "is_chr_prefixed" not in locals():
        is_chr_prefixed = str(df.iloc[0, 0]).startswith("chr")

    return is_chr_prefixed


def load_variants_file(variants):
    """
    Load variants file

    Args:
        variants (str): variants file
    """

    variants_df = pd.read_table(variants, dtype={"chrom": str, "pos": int, "ref": str, "alt": str})

    return variants_df


def annotate(variants_df, annotation_vcf, columns, match_gene, gene_mapping):
    """
    Annotate the variants file with the informations from VCF

    Args:
        variants_df (pd.DataFrame): variants file
        annotation_vcf (VariantFile): VCF file containing the annotations
        columns (list): info fields to retrieve from the VCF

    """

    logger = logging.getLogger("logger")

    # We keep track of the columns already in the variant file
    existing_columns = list(variants_df.columns)

    # We loop through each gene in the variant file
    list_annotated_df = list()
    for gene_id, gene_df in variants_df.groupby("gene_id"):

        logger.info(f"Annotating {gene_id}")

        # Annotate it with the informations from the VCF
        annotated_gene_df = annotate_gene(gene_id, gene_df, annotation_vcf, match_gene, gene_mapping, columns)

        # Add only the informations wanted by the user
        list_annotated_df.append(annotated_gene_df[existing_columns + columns])

    # Combine results for all genes
    annotated_df = pd.concat(list_annotated_df)

    return annotated_df


def annotate_gene(gene_id, gene_df, annotation_vcf, match_gene, gene_mapping, columns):
    """
    Annotate the variants from a given gene

    Args:
        gene_id (str): gene identifier
        gene_df (pd.DataFrame): variant file subsetted to a given gene
        annotation_vcf (VariantFile): VCF file containing the annotations

    """

    logger = logging.getLogger("logger")

    # We retrieve the annotation in the VCF for this gene
    gene_annotation_df = retrieve_annotation(gene_id, gene_df, annotation_vcf, match_gene, gene_mapping)

    # We merge it with the variant file
    if not gene_annotation_df.empty:
        gene_annotated_df = gene_df.merge(gene_annotation_df, how="left", on=["chrom", "pos", "ref", "alt"])
        assert gene_df.shape[0] == gene_annotated_df.shape[0]
    else:
        logger.warning(f"No annotations for gene {gene_id}")

        gene_annotated_df = gene_df
        for column in columns:
            gene_annotated_df[column] = "."

    return gene_annotated_df


def retrieve_annotation(gene_id, gene_df, annotation_vcf, match_gene, gene_mapping):
    """
    Retrieve annotation from the VCF file

    Args:
        gene_df (pd.DataFrame): variant file subsetted to a given gene
        annotation_vcf (VariantFile): VCF file containing the annotations

    """

    gene_chrom = str(gene_df.chrom.values[0]).replace("chr", "")

    # Split each gene in contiguous block (i.e. exons) and retrieve annotations
    list_block_df = list()
    for _, block in groupby(sorted(set(gene_df["pos"])), key=lambda n, c=count(): n - next(c)):

        # Get the coordinates of the current block
        start, end = as_range(block)
        # Retrieve annotation for the current block
        try:
            block_df = load_annotation(annotation_vcf, gene_chrom, start - 1, end)
            list_block_df.append(block_df)
        except ValueError:
            continue

    gene_annotations_df = format_gene_annotation(gene_id, list_block_df, match_gene, gene_mapping)

    return gene_annotations_df


def format_gene_annotation(gene_id, list_block_df, match_gene, gene_mapping):
    """
    Format the gene annotation data frame

    Args:
        gene_id (_type_): _description_
        list_block_df (_type_): _description_
        match_gene (_type_): _description_
        gff (_type_): _description_
    """

    if len(list_block_df) == 0:
        return pd.DataFrame()

    # Concatenate into gene wide annotation
    gene_annotations_df = pd.concat(list_block_df)

    if gene_annotations_df.empty:
        return pd.DataFrame()

    # Add the chr prefix if found in the variants file
    if is_chr_prefixed_variants:
        gene_annotations_df["chrom"] = ["chr" + x for x in gene_annotations_df["chrom"]]

    # Keep only annotations that match the gene identifier
    # TODO : there are a few caveats. It will work only if the variants file contains an ENSEMBL identifier
    # as gene_id and if the VCF contains either ensembl id or gene symbol
    if match_gene:

        if gene_content == "ensembl_id":
            # TODO : handle ensembl gene version
            gene_annotations_df = gene_annotations_df.loc[gene_annotations_df[gene_field] == gene_id]
        else:
            gene_symbol = gene_mapping[gene_id]
            gene_annotations_df = gene_annotations_df.loc[gene_annotations_df[gene_field] == gene_symbol]

    # TODO In some cases we have one record per transcript, for now we just pick the first record but ideally we could match on transcript
    gene_annotations_df = gene_annotations_df.loc[~gene_annotations_df[["pos", "ref", "alt"]].duplicated()]

    return gene_annotations_df


def as_range(region):
    """
    Returns genomic region boundaries

    Args:
        region (list): genomic region positions

    Returns:
        tuple: beginning and terminating position of the genomic region
    """
    l = list(region)
    return l[0], l[-1]


def load_annotation(annotation_vcf, chrom, start, end):
    """
    Access to specific region in annotation file
    Args:
        annotation_vcf (pysam.VariantFile): VCF file containing variant annotations
        chrom (str): chromosome identifier
        start (int): beginning position of the region
        end (int): terminating position of the region
    """

    list_records = list()
    for record in annotation_vcf.fetch(chrom, start, end):
        list_records.append(parse(record))

    records_df = pd.DataFrame(list_records)

    # TODO : handle annotation duplication (especially in the context of gene/transcript id)
    # records_df.drop_duplicates(inplace=True)

    return records_df


def parse(record):
    """
    Turns the VCF record into a dictionnary

    Args:
        record (VCFRecord): VCF record

    """

    record_dict = {"chrom": record.chrom, "pos": record.pos, "ref": record.ref, "alt": record.alts[0]}
    record_dict = record_dict | dict(record.info)
    return record_dict


def detect_gene_field_vcf(annotation):
    """
    Because we may want to match variants on gene identifiers, we need to retrieve
    which field in the VCF stores the gene identifier/symbol.

    Args:
        annotation (VariantFile): variant file
    """

    logger = logging.getLogger("logger")

    for i, record in enumerate(annotation):
        break

    record_info = dict(record.info)

    gene_fields = list(set(["gene", "GENE"]) & set(record_info.keys()))

    if gene_fields:

        if len(gene_fields) > 1:
            logger.warning("Several fields can contain the gene information. Matching variants on coordinates only")
        else:
            gene_field = gene_fields[0]
            if record_info[gene_field].startswith("ENSG"):
                gene_content = "ensembl_id"
            else:
                gene_content = "symbol"
    else:
        gene_field = ""
        gene_content = ""
        logger.warning("Could not find a gene field in the VCF INFO fields. Matching variants on coordinates only")

    return gene_field, gene_content


def build_gene_mapping_from_gff(gff_db):
    """
    Build mapping between gene identifier (ENSEMBL) and symbol.
    #TODO : handle cases with multiple name/ids

    Args:
        gff_db (gffutils.db): GFF database.

    """

    gene_mapping = dict()
    for gene in gff_db.all_features(featuretype="gene"):

        gene_id = gene.attributes["gene_id"][0].split(".")[0]
        gene_name = gene.attributes["gene_name"][0]

        gene_mapping[gene_id] = gene_name

    return gene_mapping


def check_columns(variants_df, annotation_vcf, columns, columns_file):
    """
    Check which column to retrieve from the VCF annotation file

    Args:
        variants_df (pd.DataFrame): variants file
        annotation_vcf (VariantFile): VCF file containing the annotations
        columns (list): list of annotation to retrieve

    """

    logger = logging.getLogger("logger")

    # If the user provided the columns to extract in a separate file we read it
    if columns_file:
        annotation_columns = read_columns_from_file(columns_file)
    # If he provided them as a string we split it
    elif columns:
        annotation_columns = columns.split(",")
    # If he did not provide any column we retrieve all the fields in INFO
    else:
        for i, record in enumerate(annotation_vcf):
            break
        annotation_columns = list(dict(record.info).keys())

    variants_columns = variants_df.columns

    shared_columns = set(variants_columns) & set(annotation_columns)

    for shared_column in shared_columns:
        logger.warning(f"{shared_column} column exists already in variant file")

    return annotation_columns


def build_gene_mapping_from_column(variants_df):
    """
    Build a dictionnary that maps gene ENSEMBL identifiers to gene symbol

    Args:
        variants_df (pd.DataFrame): variants file
    """

    df = variants_df[["gene_id", "symbol"]]
    df = df.loc[~df[["gene_id"]].duplicated()]

    gene_mapping = dict(zip(df.gene_id, df.symbol))

    return gene_mapping


@click.command()
@click.argument("variants", type=click.Path(exists=True))
@click.argument("annotation", type=click.Path(exists=True))
@click.argument("output", type=click.Path())
@click.option(
    "-c", "--columns", default="", type=str, help="Columns to use in the annotation file. Should be comma separated"
)
@click.option(
    "-C", "--columns-file", default="", type=str, help="File listing columns to extract from the annotation file"
)
@click.option("--match-gene", is_flag=True, help="Match the annotation based on gene identifier or symbol")
@click.option(
    "--gff",
    type=str,
    help="GFF file used to build the rates file, and that contains the ENSEMBL identifier / gene symbol matching",
)
def cli(variants, annotation, output, columns, columns_file, match_gene, gff):
    """
    Annotate a variants file (e.g. rates file) with informations stored in a VCF.
    CAVEAT : For now the script works only with variant file using ENSEMBL gene identifiers, and with VCF
    such as the popeve and Cosmic one. There is no guarantee that it will work with other data.

    Args:
        variants (str): variant file
        annotation (str): annotation to add to the variant file in VCF format
        output (str): annotated variant file
        columns (str): columns to retrieve from the VCF. Should be comma separated
        columns_file (str): File listing which annotations to extract from dbNSFP
        match_gene (boolean): whether to match annotation using coordinates only or gene id
        gff (str): gff file used to build the rates file, contains matching between gene identifier and gene symbol that
        can be used if the annotation contains only gene symbol
    """

    init_log()

    # Handle differences in chromosome representations
    global is_chr_prefixed_variants, is_chr_prefixed_annotation
    is_chr_prefixed_variants = is_chr_prefixed(variants)
    is_chr_prefixed_annotation = is_chr_prefixed(annotation)

    # Load variants file
    variants_df = load_variants_file(variants)

    # Load indexed annotation file
    annotation_vcf = pysam.VariantFile(annotation)

    # See if we can retrieve gene informations in the VCF
    global gene_field, gene_content
    gene_field, gene_content = detect_gene_field_vcf(annotation_vcf)
    if gene_field == "":
        match_gene = False

    # Some annotations rely on gene symbol rather than identifiers, we therefore build a mapping dictionnary
    # either from the gff file that was used to generate the rates file
    if gff:
        gff_db = load_gff(gff, "gff.db")
        gene_mapping = build_gene_mapping_from_gff(gff_db)
    # or from a symbol column if found in the file
    elif "symbol" in variants_df.columns:
        gene_mapping = build_gene_mapping_from_column(variants_df)
    else:
        gene_mapping = dict()

    # Annotate variants file
    annotation_columns = check_columns(variants_df, annotation_vcf, columns, columns_file)
    annotated_df = annotate(variants_df, annotation_vcf, annotation_columns, match_gene, gene_mapping)

    # Export annotated file
    annotated_df.to_csv(output, sep="\t", index=False, na_rep=".")


if __name__ == "__main__":
    cli()



import os
import gffutils
import logging
import yaml


def read_columns_from_file(columns_file):
    """
    Read columns to extract from annotation file

    Args:
        columns_file (str): file listing columns to extract from TSV/VCF/dbNSFP
    """

    with open(columns_file, "r") as f:
        columns = [x.strip() for x in f.readlines()]

    return columns


def load_gff(gff_file, gff_db_out=""):
    """
    Create or load a gffutils database.

    Args:
        gff_file (str): Path to a GFF or a gffutils database file.
        gff_db_out (str): If a GFF file is provided, the user can name the yet to be created gffutils db file.
    Returns:
        gffutils.db: GFF database.

    """

    logger = logging.getLogger("logger")

    # gffutils db input
    if gff_file.endswith(".db"):
        logger.info(f"Loading gffutils database {gff_file}")
        gff_db = gffutils.FeatureDB(gff_file)
    # GFF input
    else:
        if gff_db_out:
            gff_db_path = gff_db_out
        else:
            gff_db_path = gff_file + ".db"
        logger.info(f"Creating GFF db {gff_db_path}")

        try:
            os.remove(gff_db_path)
            logger.info(f"Removed old gffutils database : {gff_db_path}")
        except OSError:
            pass

        logger.info(f"Creating gffutils database : {gff_db_path}")
        gff_db = gffutils.create_db(gff_file, gff_db_path, merge_strategy="create_unique")

    return gff_db


def load_conf(filename):
    """Load the content of a YAML configuration file in a dictionnary

    Args:
        filename (str): Path to a YAML configuration file.

    Returns:
        dict: Configuration dictionnary.
    """

    with open(filename) as file:
        conf = yaml.load(file, Loader=yaml.FullLoader)

    return conf


def superseed_conf(conf, command_params):
    """Superseed configuration in config file with parameters from the command line

    Args:
        conf (dict): configuration from the config file (if any)
        command_params (dict): command line parameters

    Returns:
        dict: Configuration dictionnary.
    """

    for key, value in command_params.items():
        if value and key != "config":
            conf[key.upper()] = value

    return conf



import pandas as pd
import numpy as np

from denovowest.utils.params import INFRAME_MISSENSE_RATIO, FRAMESHIFT_NONSENSE_RATIO


def prepare_scores(dnm_df, rates_df, score_column, runtype, impute_missing=False):
    """
    Assign scores to variants in the DNM and rates file.
    The score used for the scores is first min-max transformed (accounts for CEP with negative scores)

    Args:
        dnm_df (pd.DataFrame): DNM dataframe
        rates_df (pd.DataFrame): rates dataframe that contains all possible SNV
        score_column (str) : CEP scores
    """

    # Infer indel scores and mutation rates
    if runtype == "ns":
        indel_rates_df = infer_indel_scores_and_rates(rates_df, score_column)
        dnm_df = assign_dnm_indel_scores(dnm_df, indel_rates_df, rates_df, score_column)

    # Impute scores for variants with missing scores
    if impute_missing:
        dnm_df, rates_df = impute_missing_scores(dnm_df, rates_df, score_column)

    # Consolidate the rates df by adding the indel rates
    if runtype == "ns":
        rates_df = pd.concat([rates_df, indel_rates_df])

    return dnm_df, rates_df


def infer_indel_scores_and_rates(rates_df, score_column):
    """
    Infer expected inframe and frameshift scores and mutation rates per gene based
    on the missense and nonsense mutations respectively.

    Args:
        rates_df (pd.DataFrame): rates dataframe
        score_column (str) : CEP scores


    Returns:
        pd.DataFrame: per-gene inframe and frameshift mutation rates and associated scores
    """

    indel_rates_list = list()
    for gene_id, gene_df in rates_df.groupby("gene_id"):

        # Get the inframe mutation rate based on the cumulative missense mutation rates, and assign a score from gene-based median missense score
        gene_inframe_rate = gene_df.loc[gene_df.consequence == "missense", "prob"].sum() * INFRAME_MISSENSE_RATIO
        gene_inframe_scores = gene_df.loc[gene_df.consequence == "missense", score_column].median()

        # Get the frameshift mutation rate based on the cumulative nonsense mutation rates, and assign a score from gene-based median nonsense score
        gene_frameshift_rate = gene_df.loc[gene_df.consequence == "nonsense", "prob"].sum() * FRAMESHIFT_NONSENSE_RATIO
        gene_frameshift_scores = gene_df.loc[gene_df.consequence == "nonsense", score_column].median()

        inframe_row = {
            "gene_id": gene_id,
            "consequence": "inframe",
            "prob": gene_inframe_rate,
            score_column: gene_inframe_scores,
        }

        frameshift_row = {
            "gene_id": gene_id,
            "consequence": "frameshift",
            "prob": gene_frameshift_rate,
            score_column: gene_frameshift_scores,
        }

        indel_rates_list.append(inframe_row)
        indel_rates_list.append(frameshift_row)

    indel_rates_df = pd.DataFrame(indel_rates_list)

    return indel_rates_df


def assign_dnm_indel_scores(dnm_df, indel_rates_df, rates_df, score_column):
    """
    Assign gene-based indel scores taken from the rates file to observed inframe
    and frameshift variants

    Args:
        dnm_df (pd.DataFrame): observed DNM
        indel_rates_df (pd.DataFrame) : per-gene score for frameshit and inframe annotated indels
        rates_df (pd.DataFrame): rates dataframe
        score_column (str) : CEP scores

    Returns:
        pd.DataFrame: observed DNM with scores associated to indels
    """

    list_scores = list()
    for _, dnm in dnm_df.iterrows():

        # If the DNM is not an indel we keep the current score
        if (len(dnm.alt) - len(dnm.ref)) == 0:
            list_scores.append(dnm[score_column])
            continue

        # If it is an indel but not annotated as inframe or frameshift, we get the
        # gene median corresponding score
        if not (dnm.consequence in ["inframe", "frameshift"]):

            generates_df = rates_df.loc[rates_df.gene_id == dnm.gene_id]
            score = generates_df.loc[generates_df.consequence == dnm.consequence, score_column].median()
            list_scores.append(score)

        # Otherwise we get the gene inframe/frameshift score
        else:
            list_scores.append(
                indel_rates_df.loc[
                    (indel_rates_df.gene_id == dnm.gene_id) & (indel_rates_df.consequence == dnm.consequence),
                    score_column,
                ].iloc[0]
            )

    dnm_df.loc[:, score_column] = list_scores
    return dnm_df


def impute_missing_scores(dnm_df, rates_df, score_column):
    """
    Some CEPs do not assign a score to each variant.
    Here we impute the missing scores looking at the median score for each type
    of consequence per gene.

    Args:
        dnm_df (pd.DataFrame): observed DNM
        rates_df (pd.DataFrame): expected mutations

    Returns:
        tuple(pd.DataFrame, pd.DataFrame): DNM and rates dataframes with imputed missing scores
    """

    # Get the median values per gene and per consequence type
    median_values_dict = dict()
    for gene_id, generates_df in rates_df.groupby("gene_id"):
        median_values_dict[gene_id] = dict()
        for consequence, generates_cq_df in generates_df.groupby("consequence"):
            median_value = generates_cq_df[score_column].median()
            median_values_dict[gene_id][consequence] = median_value

    # Impute missing scores on rates dataframe
    imputed_rates_scores = list()
    for _, variant in rates_df.iterrows():
        if np.isnan(variant[score_column]):
            imputed_rates_scores.append(median_values_dict[variant.gene_id][variant.consequence])
        else:
            imputed_rates_scores.append(variant[score_column])
    rates_df["score_before_imputation"] = rates_df[score_column]
    rates_df[score_column] = imputed_rates_scores

    # Impute missing scores on DNM dataframe
    imputed_dnm_scores = list()
    for _, variant in dnm_df.iterrows():
        if np.isnan(variant[score_column]):
            imputed_dnm_scores.append(median_values_dict[variant.gene_id][variant.consequence])
        else:
            imputed_dnm_scores.append(variant[score_column])

    dnm_df["score_before_imputation"] = dnm_df[score_column]
    dnm_df[score_column] = imputed_dnm_scores

    return dnm_df, rates_df



#!/usr/bin/env python
import pandas as pd
import click
import sys
import logging

from denovowest.utils.log import init_log


@click.command()
@click.argument("rates")
@click.argument("shet")
@click.argument("output")
@click.option("--threshold", default=0.15, help="shet threshold to determine if gene is high or low")
def annotate_shet(rates, shet, output, threshold):
    """Adds shet boolean to rates file

    Args:
        rates (str): Path to rates file
        shet (str): Path to shet file
        output (str): Path to output file (merged dataframe)
    """

    init_log()
    logger = logging.getLogger("logger")

    # Load rates file
    rates_df = pd.read_table(rates)
    if rates_df.empty:
        logger.warning("Rates file is empty")
        rates_df["shethigh"] = None
        rates_df.to_csv(output, sep="\t", index=False)
        sys.exit(0)
    rates_df["shethigh"] = False

    # Load shet file
    shet_df = pd.read_table(shet, index_col=1)

    # For each gene
    nb_gene_not_in_shet = 0
    for gene_id_orig, gene_rates_df in rates_df.groupby("gene_id"):
        # Remove gene version if any
        gene_id = gene_id_orig.split(".")[0]
        gene_id = gene_id.replace("gene:", "")

        try:
            mean_shet = shet_df.loc[gene_id, "mean_s_het"]
            if mean_shet > threshold:
                rates_df.loc[rates_df["gene_id"] == gene_id_orig, "shethigh"] = True

        except KeyError:
            nb_gene_not_in_shet += 1

    logger.info(f"{nb_gene_not_in_shet}/{len(set(rates_df['gene_id']))} genes from rates file not found in shet")

    # Export results
    rates_df.to_csv(output, sep="\t", index=False)


if __name__ == "__main__":
    merged_df = annotate_shet()



#!/usr/bin/env python
import pandas as pd
import click
import pysam
import sys
from itertools import groupby, count
import utils
import logging
import gzip

is_chr_prefixed_variants = False
is_chr_prefixed_annotation = False

# TODO : Handle annotation with matching on gene identifier/symbol


def is_chr_prefixed(df_path):
    """
    Chromosomes can be defined as "N" or "chrN"
    If it differs between the input file and the annotation file it prevents records matching

    Args:
        df_path (str): variants or annotation file

    Returns:
        bool: whether the chromosomes are prefixed or not in {df_path}
    """

    df = pd.read_csv(df_path, sep="\t", nrows=100)

    for key in ["chrom", "chr", "#chr", "#chrom"]:
        if key in df.columns:
            is_chr_prefixed = str(df[key].iloc[0]).startswith("chr")
            break

    # TODO : Better handle this case
    if "is_chr_prefixed" not in locals():
        is_chr_prefixed = str(df.iloc[0, 0]).startswith("chr")

    return is_chr_prefixed


def load_variants_file(variants):
    """
    Load variants file (e.g. DNM/rates file)

    Args:
        variants (str): variants file
    """

    variants_df = pd.read_table(variants, dtype={"chrom": str, "pos": int, "ref": str, "alt": str})

    return variants_df


def annotate(variants_df, annotation_df, columns_indices, columns_names):
    """
    Annotate variants file with the selected columns from the annotation file.

    Args:
        variants_df (pd.DataFrame): variants
        annotation_df (pd.DataFrame): annotations
        columns_indices (list):  index of annotations to extract
        columns_names (list):  names of annotations to extract
    """

    list_annotated_df = list()
    for gene_id, gene_df in variants_df.groupby("gene_id"):

        annotated_gene_df = annotate_gene(gene_id, gene_df, annotation_df, columns_indices)
        list_annotated_df.append(annotated_gene_df)

    # Combine results for all genes
    annotated_df = pd.concat(list_annotated_df)
    annotated_df.rename(dict(zip([str(x) for x in columns_indices], columns_names)), axis=1, inplace=True)

    return annotated_df


def annotate_gene(gene_id, gene_df, annotation_df, columns):
    """
    Annotate the current gene

    Args:
        gene_id (str): current gene identifier
        gene_df (pd.DataFrame): variants found in the current gene
        annotation_df (pd.DataFrame): annotations
        columns (list):  indices of annotation to retrieve

    """

    logger = logging.getLogger("logger")

    gene_annotation_df = retrieve_annotation(gene_df, annotation_df, columns, gene_id)

    if not gene_annotation_df.empty:
        gene_annotated_df = gene_df.merge(gene_annotation_df, how="left", on=["chrom", "pos", "ref", "alt"])
    else:
        logger.warning(f"No annotations for gene {gene_id}")
        gene_annotated_df = gene_df

    assert gene_df.shape[0] == gene_annotated_df.shape[0]

    # for column in columns:
    #     if column not in gene_annotated_df.columns:
    #         gene_annotated_df[column] = "."

    return gene_annotated_df


def retrieve_annotation(gene_df, annotation_df, columns, gene_id):
    """
    Retrieve annoations for the current gene

    Args:
        gene_df (pd.DataFrame): _description_
        annotation_df (pd.DataFrame): _description_
        columns (list):  indices of annotation to retrieve
        gene_id (str) : gene identifier

    """

    logger = logging.getLogger("logger")

    gene_chrom = str(gene_df.chrom.values[0]).replace("chr", "")
    # Split each gene in contiguous block (i.e. exons) and retrieve annotations
    list_block_df = list()
    for _, block in groupby(sorted(set(gene_df["pos"])), key=lambda n, c=count(): n - next(c)):

        # Get the coordinates of the current block
        start, end = as_range(block)

        # Retrieve annotation for the current block
        try:
            block_cadd_df = load_annotation(annotation_df, gene_chrom, start - 1, end, columns)
            list_block_df.append(block_cadd_df)
        except ValueError:
            continue

    # Concatenate into gene wide annotation
    if list_block_df:
        gene_annotations_df = pd.concat(list_block_df)

        # Add the chr prefix if found in the variants file
        if not gene_annotations_df.empty and is_chr_prefixed_variants:
            gene_annotations_df["chrom"] = ["chr" + x for x in gene_annotations_df["chrom"]]
    else:
        gene_annotations_df = pd.DataFrame()

    # Remove duplicated rows if any
    if not gene_annotations_df.empty:
        nb_rows_before_duplication = gene_annotations_df.shape[0]
        gene_annotations_df = gene_annotations_df.loc[~gene_annotations_df[["chrom", "pos", "ref", "alt"]].duplicated()]
        if nb_rows_before_duplication != gene_annotations_df.shape[0]:
            logger.warning(f"There were multiple annotation available for {gene_id}")

    return gene_annotations_df


def as_range(region):
    """
    Returns genomic region boundaries

    Args:
        region (list): genomic region positions

    Returns:
        tuple: beginning and terminating position of the genomic region
    """
    l = list(region)
    return l[0], l[-1]


def load_annotation(annotation_file, chrom, start, end, columns):
    """
    Access to specific region in annotation file
    Args:
        annotation_file (pysam.TabixFile): annotation file
        chrom (str): chromosome identifier
        start (int): beginning position of the region
        end (int): terminating position of the region
    """

    list_records = list()
    for record in annotation_file.fetch(chrom, start, end, columns):
        list_records.append(parse(record, columns))

    records_df = pd.DataFrame(list_records)

    # TODO : handle annotation duplication (especially in the context of gene/transcript id)
    records_df.drop_duplicates(inplace=True)

    return records_df


def parse(line, columns):

    record = line.split("\t")

    record_dict = dict()
    record_dict["chrom"] = record[0]
    record_dict["pos"] = int(record[1])
    record_dict["ref"] = record[2]
    record_dict["alt"] = record[3]

    for column in columns:
        record_dict[str(column)] = record[column]

    return record_dict


def extract_columns_indices(annotation_df, columns, columns_file):
    """
    Extract the indices of the column to use from the annotation file

    Args:
        annotation_df (pysam.TabixFile): annotation indexed file
        columns (str): columns to extract from the annotation file, comma separated
        columns_file (str)
    """

    logger = logging.getLogger("logger")

    try:
        all_columns = annotation_df.header[0].split("\t")
    except IndexError:
        # If the annotation file header does not start with #, the Tabix header will be empty
        # and we need to retrieve it another way
        with gzip.open(annotation_df.filename, "rt") as f:
            all_columns = f.readline().strip().split("\t")

    # If the user provided the columns to extract in a separate file we read it
    if columns_file:
        columns_to_extract = utils.read_columns_from_file(columns_file)
    # If he provided them as a string we split it
    elif columns:
        columns_to_extract = columns.split(",")
    # If he did not provide any column we retrieve all non standard columns
    else:
        columns_to_extract = all_columns[4:]

    # Retrieve indices of columns to extract
    columns_indices = list()
    columns_names = list()
    for idx, name in enumerate(all_columns):
        if name in columns_to_extract:
            columns_indices.append(idx)
            columns_names.append(name)

    # If a score can not be retrieved from dbNSFP we stop here
    columns_not_found = set(columns_to_extract) - set(columns_names)
    if columns_not_found:
        logger.error(f"The following columns could not be retrieved from the annotation file : {columns_not_found}")
        sys.exit(1)

    return columns_indices, columns_names


def check_columns(input_columns, custom_columns):
    """
    Check whether some of the columns wanted by the user are already in the input file

    Args:
        input_columns (str): input data frame columns
        custom_columns (str): annotation columns

    """

    logger = logging.getLogger("logger")

    shared_columns = set(input_columns) & set(custom_columns)

    for shared_column in shared_columns:
        logger.warning(
            f"{shared_column} column exists already in variant file. The one coming from the custom annotation file will be suffixed with _custom"
        )

    return shared_columns


@click.command()
@click.argument("variants", type=click.Path(exists=True))
@click.argument("annotation", type=click.Path(exists=True))
@click.argument("output", type=click.Path())
@click.option(
    "-c", "--columns", default="", type=str, help="Columns to use in the annotation file. Should be comma separated"
)
@click.option(
    "-C", "--columns-file", default="", type=str, help="File listing columns to extract from the annotation file"
)
def cli(variants, annotation, output, columns, columns_file):
    """
    Annotate a TSV variants file (DNM or rates) with information coming from a
    TAB-delimited genome position file indexed with tabix
    (https://www.htslib.org/doc/tabix.html)

    Args:
        variants (str): Variants file (e.g. rates, DNM) starting with the following columns [gene_id,chrom,pos,ref,alt]
        annotation (str): annotation file
        output (str): variants file annotated
        columns (str): columns to use in the annotation file (comma separated)
        columns_file (str): File listing which annotations to extract from dbNSFP

    """

    utils.init_log()

    global is_chr_prefixed_variants, is_chr_prefixed_annotation
    is_chr_prefixed_variants = is_chr_prefixed(variants)
    is_chr_prefixed_annotation = is_chr_prefixed(annotation)

    # Load variants file
    variants_df = load_variants_file(variants)

    # Load indexed annotation file
    annotation_df = pysam.TabixFile(annotation)

    # Get colums
    columns_indices, columns_names = extract_columns_indices(annotation_df, columns, columns_file)

    # If some columns are already found in the input file, we prefix the new ones with "dbnsfp"
    existing_columns = check_columns(variants_df.columns, columns_names)
    columns_names = [f"{col}_custom" if col in existing_columns else col for col in columns_names]

    # Annotate variants file
    annotated_df = annotate(variants_df, annotation_df, columns_indices, columns_names)

    # Export annotated file
    annotated_df.to_csv(output, sep="\t", index=False, na_rep=".")


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
import click
import pandas as pd
import logging
import re

from denovowest.utils.log import init_log
from denovowest.utils.params import CDS_OFFSET
from denovowest.utils.io_helpers import load_gff


@click.command()
@click.argument("dnm")
@click.argument("gene_list")
@click.option("--output_kept_dnm")
@click.option("--output_discarded_dnm")
@click.option("--gff")
def filter_dnm(dnm, gene_list, output_kept_dnm, output_discarded_dnm, gff):
    """
    Remove from DNM table the variants in genes not found in gene list.
    The gene list has been either provided by the user or built from the rates file.

    If the GFF option is provided, variants not found in CDS regions will be discarded.

    Args:
        dnm (str): path to DNM file
        gene_list (str): path to gene list file
        output_kept_dnm (str): path to kept DNM table
        output_discarded_dnm (str): path to filtered DNM table
        gff (str): path to GFF file
    """

    genes = load_gene_list(gene_list)
    dnm_df = load_dnm(dnm, genes)

    # Filter DNM table on gene list
    dnm_df, dnm_discarded_df = filter_on_gene_list(dnm_df, genes)

    # Filter DNM table on CDS regions
    if gff:
        dnm_df, dnm_discarded_gff_df = filter_on_gff(dnm_df, gff)
        dnm_discarded_df = pd.concat([dnm_discarded_df, dnm_discarded_gff_df])

    # Export DNM tables
    dnm_df.to_csv(output_kept_dnm, sep="\t", index=False)
    dnm_discarded_df.to_csv(output_discarded_dnm, sep="\t", index=False)

    # Log stats
    log_stats(dnm_df, dnm_discarded_df, output_discarded_dnm)


def filter_on_gene_list(dnm_df, gene_list):
    """
    Remove from DNM table the variants in genes not found in gene list

    Args:
        dnm_df (pd.DataFrame): DNM table
        gene_list (list): gene list
    """

    dnm_kept_df = dnm_df.loc[dnm_df["gene_id"].isin(gene_list)]
    dnm_discarded_df = dnm_df.loc[~dnm_df["gene_id"].isin(gene_list)].copy()
    dnm_discarded_df["reason"] = "not_in_gene_list"

    return dnm_kept_df, dnm_discarded_df


def filter_on_gff(dnm_df, gff):
    """
    Remove from DNM table the variants not found in CDS regions

    Args:
        dnm_df (pd.DataFrame): DNM table
        gff (str): path to GFF file or gffutils database
    """

    gff_db = load_gff(gff)

    idx_not_in_cds = []
    for idx, dnm in dnm_df.iterrows():
        if not is_in_cds(gff_db, dnm):
            idx_not_in_cds.append(idx)

    dnm_discarded_df = dnm_df.loc[idx_not_in_cds].copy()
    dnm_discarded_df["reason"] = "not_in_cds"
    dnm_kept_df = dnm_df.drop(idx_not_in_cds)

    return dnm_kept_df, dnm_discarded_df


def is_in_cds(gff_db, dnm):
    """
    Check if a given DNM is in a CDS region defined in the GFF file

    Args:
        gff_db (gffutils.FeatureDB): gffutils database
        dnm (pd.Series): DNM row
    """

    pos = dnm["pos"]
    gene_id = dnm["gene_id"]

    gene = gff_db[gene_id]
    for transcript in gff_db.children(gene, level=1):
        for cds in gff_db.children(transcript, featuretype="CDS", order_by="start"):
            if (pos >= cds.start - CDS_OFFSET) and (pos <= cds.end + CDS_OFFSET):
                return True

    return False


def load_dnm(dnm, genes):
    """
    Load DNM file

    Args:
        dnm (str): path to DNM file
        genes (list) : list of gene identifiers
    """

    dnm_df = pd.read_csv(dnm, sep="\t")
    dnm_df["gene_id"] = format_gene_id(list(dnm_df["gene_id"]), genes)

    return dnm_df


def load_gene_list(gene_list):
    """
    Load gene list that will be used to filter the DNM table

    Args:
        gene_list (str): path to gene list
    """

    with open(gene_list, "r") as f:
        genes = f.readlines()

    genes = [gene_id.strip() for gene_id in genes]

    return genes


def format_gene_id(genes_in_dnm, genes_list):
    """
    Get rid of trailing spaces after gene ids and matches gene identifiers from rates file

    Args:
        genes_in_dnm (list): list of genes in DNM file
        genes_list (list) : list of genes provided by the user or taken from the rates file
    """

    # Strip genes identifiers to remove typo mistakes
    genes_in_dnm = [gene_id.strip() for gene_id in genes_in_dnm]

    # If the DNM file gene identifiers do not contain version number but the rates file does, we suffix the gene identifier with the version
    pattern_ensembl_version = r"^ENSG\d{11}\.\d+$"
    pattern_ensembl_id_only = r"^ENSG\d{11}$"
    if bool(re.match(pattern_ensembl_id_only, genes_in_dnm[0])) & bool(
        re.match(pattern_ensembl_version, genes_list[0])
    ):
        genes_dict = {x.split(".")[0]: x for x in genes_list}
        genes_in_dnm = [genes_dict[x] if x in genes_dict.keys() else x for x in genes_in_dnm]

    return genes_in_dnm


def log_stats(dnm_kept_df, dnm_discarded_df, output_discarded_dnm):
    """
    Log stats about the DNM table filtering

    Args:
        dnm_kept_df (pd.DataFrame): DNM kept
        dnm_discarded_df (pd.DataFrame): DNM filtered out
        output_discarded_dnm (str): path to filtered DNM table
    """
    nb_dnm = dnm_kept_df.shape[0] + dnm_discarded_df.shape[0]
    logger = logging.getLogger("logger")

    if dnm_discarded_df.empty:
        logger.info(f"All DNM ({nb_dnm}) were retained")
    else:
        nb_dnm_discarded = dnm_discarded_df.shape[0]
        logger.warning(f"{nb_dnm_discarded}/{nb_dnm} DNM were discarded, of which :")
        for reason, count in dict(dnm_discarded_df.reason.value_counts()).items():
            logger.warning(f"- {reason} : {count} DNM")
        logger.warning(f"Check filtered DNM table : {output_discarded_dnm}")


if __name__ == "__main__":
    init_log()
    filter_dnm()



#!/usr/bin/env python
import pandas as pd
import click
import pysam
import sys
import logging

from itertools import groupby, count
from denovowest.utils.log import init_log
from denovowest.utils.io_helpers import read_columns_from_file, load_gff


# TODO : Better handle the ENSG versions
# TODO : See if we could improve the selection when multiple records exist for a same variant (@remove_duplicates)


def load_dbnsfp(
    dbnsfp_file, chrom, start, end, columns_indices, gene_id, gff_db=None, ensembl_gene_id_map_version=None
):
    """
    Access to specific region in an dbNSFP indexed file
    Args:
        annotation_file (pysam.TabixFile): Annotation file containing CEP scores
        chrom (str): chromosome identifier
        start (int): beginning position of the region
        end (int): terminating position of the region
        columns_indices(list) : indices of the dbNSFP columns to retrieve
        gene_id(str) : gene identifier used in DNM and rates file
        gff_db (gffutils.FeatureDB): gffutils database
        ensembl_gene_id_map_version (dict) : maps ENSG with version to ENSG without version (e.g. {ENSG00000010404 : ENSG00000010404.1})
    """

    # Retrieve the transcripts associated to the genes from the GFF file
    if gff_db:
        transcript_ids = get_transcripts(gene_id, gff_db, ensembl_gene_id_map_version)
    else:
        transcript_ids = list()

    list_annotated_records = list()
    for record in dbnsfp_file.fetch(chrom, start, end):
        record_dict = parse(record, columns_indices, gene_id, transcript_ids)
        if record_dict:
            list_annotated_records.append(record_dict)

    return pd.DataFrame(list_annotated_records)


def get_transcripts(gene_id, gff_db, ensembl_gene_id_map_version):
    """
    Return transcripts identifiers of the transcripts found in the GFF file for a given gene

    Args:
        gene_id (str): gene identifier
        gff_db (gffutils.FeatureDB): gffutils database
        ensembl_gene_id_map_version (dict) : maps ENSG with version to ENSG without version (e.g. {ENSG00000010404 : ENSG00000010404.1})

    Returns:
        list: list of transcripts identifiers
    """

    if "." in gene_id:
        gene = gff_db[gene_id]
    else:
        gene = gff_db[ensembl_gene_id_map_version[gene_id]]

    list_transcript_ids = list()
    for transcript in gff_db.children(gene, featuretype="transcript", order_by="start"):
        list_transcript_ids += [x.split(".")[0] for x in transcript["transcript_id"]]

    return list_transcript_ids


def parse(record, columns_indices, gene_id, transcript_ids_gff=list()):
    """
    Parse a dbNSFP record (line) and extract the data of interest

    Args:
        record (str): line in dbNSFP
        columns_indices (list): indices of columns to retrieve from the dbNSFP records
        gene_id(str) : ENSEMBL gene identifier from the input file
        transcript_ids_gff (list) : list of {gene_id} transcripts found in the GFF file (if provided)

    Returns:
        dict: informations about a mutation (pos, CEP scores)
    """

    record = record.split("\t")

    record_dict = dict()
    record_dict["chrom"] = record[0]
    record_dict["pos"] = int(record[1])
    record_dict["ref"] = record[2]
    record_dict["alt"] = record[3]

    # TODO : should retrieve the index rather than relying ona position that might change with versions of dbNSDP
    ensembl_gene_ids_dbnsfp = record[13].split(";")
    transcript_ids_dbnsfp = record[14].split(";")

    # Check the intersection between the transcripts from the user gff and dbNSFP
    if len(set(transcript_ids_gff) & set(transcript_ids_dbnsfp)) != 0:
        transcript_in_dbnsfp = True
    else:
        transcript_in_dbnsfp = False
    record_dict["transcript_in_dbnsfp"] = transcript_in_dbnsfp

    # If the record is associated to an overlapping gene, skip it
    if gene_id.split(".")[0] not in ensembl_gene_ids_dbnsfp:
        return dict()

    # Retrieve the scores
    for column_index in columns_indices:
        record_dict[str(column_index)] = record[column_index]

    # If a GFF has been provided, retrieve the maximum transcript-based score
    if transcript_ids_gff:

        for column_index in columns_indices:
            list_scores = record[column_index].split(";")

            score = "."

            # If there is only one score, at this point we assume it is a per-gene score
            if len(list_scores) == 1:
                score = list_scores[0]
                record_dict[str(column_index)] = score
                continue

            # We first try to retrieve the transcript based score
            if transcript_in_dbnsfp:
                score = get_transcript_based_score(list_scores, transcript_ids_gff, transcript_ids_dbnsfp)

            # Some MANE transcripts are missing in dbNSFP, therefore if no score associated to a transcript could be retrieved
            # we switch to gene mode
            if score == ".":
                score = get_gene_based_score(list_scores, gene_id.split(".")[0], ensembl_gene_ids_dbnsfp)

            record_dict[str(column_index)] = score

    return record_dict


def get_transcript_based_score(list_scores, transcript_ids_gff, transcript_ids_dbnsfp):
    """
    Return the maximum transcript-based score among transcripts found in the GFF

    Args:
        list_scores (list): list of transcript-based scores for a given metric
        transcript_ids_gff(list) : list of transcript identifiers found in the GFF
        transcript_ids_dbnsfp(list) : list of transcript identifiers in dbNSFP for the given variant
    """

    list_idx = list()
    for idx, transcript_id in enumerate(transcript_ids_dbnsfp):
        if transcript_id in transcript_ids_gff:
            list_idx.append(idx)

    # Not all dbNSFP columns contain scores, and they do not necessarily match the number of values in gene/transcript columns
    try:
        max_score = max_value([list_scores[i] for i in list_idx])
    except IndexError:
        # In that case we just return the whole value
        max_score = ";".join(list_scores)

    return max_score


def get_gene_based_score(list_scores, gene_id, gene_ids_dbnsfp):
    """
    Return the maximum gene-based score among all transcripts associated to this gene

    Args:
        list_scores (list): list of transcript-based scores for a given metric
        transcript_ids_gff(list) : list of transcript identifiers found in the GFF
        transcript_ids_dbnsfp(list) : list of transcript identifiers in dbNSFP for the given variant
    """

    list_idx = list()
    for idx, dbnsfp_gene_id in enumerate(gene_ids_dbnsfp):
        if dbnsfp_gene_id == gene_id:
            list_idx.append(idx)

    # Not all dbNSFP columns contain scores, and they do not necessarily match the number of values in gene/transcript columns
    try:
        max_score = max_value([list_scores[i] for i in list_idx])
    except IndexError:
        # In that case we just return the whole value
        max_score = ";".join(list_scores)

    return max_score


def max_value(list_scores):
    """
    Return the maximum value in a list that can contains "."

    Args:
        list_scores (list): list of transcript-based scores for a given metric

    Returns:
        float: maximum score in the list
    """
    try:
        max_score = max([float(x) for x in list_scores if x != "."])
    except ValueError:

        # Not all dbNSFP columns contain scores, and they do not necessarily match the number of values in gene/transcript columns
        if set(list_scores) == {"."}:
            # If there is no score we return "."
            max_score = "."
        else:
            # If it was not a score column we return the whole value
            max_score = ";".join(list_scores)

    return max_score


def as_range(region):
    """
    Returns genomic region boundaries

    Args:
        region (list): genomic region positions

    Returns:
        tuple: beginning and terminating position of the genomic region
    """

    return region[0], region[-1]


def extract_columns_from_dbNFP(dbnsfp, columns, columns_file):
    """
    Extract the indices of the column to use in dbnsfp

    Args:
        dbnsfp (pysam.TabixFile): dbNSFP indexed file
        columns(str) : Annotations to extract form dbNSFP (comma separated)
        columns_file (str): File listing which annotations to extract from dbNSFP
    """

    logger = logging.getLogger("logger")

    # If the user provided the columns to extract in a separate file we read it
    if columns_file:
        columns_to_extract = read_columns_from_file(columns_file)
    # If he provided them as a string we split it
    elif columns:
        columns_to_extract = columns.split(",")
    # If he did not provide any column we retrieve all columns in dbNSFP
    else:
        columns_to_extract = dbnsfp.header[0].split("\t")[4:]

    # Retrieve dbNSFP columns
    dbnsfp_columns = dbnsfp.header[0].split("\t")

    # Find the index of all columns asked by the user
    list_indices = list()
    found_column = list()
    for idx, name in enumerate(dbnsfp_columns):
        if name in columns_to_extract:
            list_indices.append(idx)
            found_column.append(name)

    # If a column can not be retrieved from dbNSFP we stop here
    columns_not_found = set(columns_to_extract) - set(found_column)
    if columns_not_found:
        logger.error(f"The following columns could not be retrieved from dbNFSP : {columns_not_found}")
        sys.exit(1)

    return list_indices, found_column


def check_columns(input_columns, dbnsfp_columns):
    """
    Check whether some of the columns wanted by the user are already in the input file

    Args:
        input_columns (str): input data frame columns
        dbnsfp_columns (str): annotation columns

    """

    logger = logging.getLogger("logger")

    shared_columns = set(input_columns) & set(dbnsfp_columns)

    for shared_column in shared_columns:
        logger.warning(
            f"{shared_column} column exists already in variant file. The one coming from dbNSFP will be suffixed with _dbnsfp"
        )

    return shared_columns


def load_rates_file(rates, output):
    """
    Load rates file

    Args:
        rates (str): path to a variant file in TSV format
        output (str): output file

    """

    logger = logging.getLogger("logger")

    # Load rates file
    rates_df = pd.read_table(rates, dtype={"chrom": str, "pos": int, "ref": str, "alt": str})

    # Edge case where when splitting the processes we end up with an empty input file
    if rates_df.empty:
        logger.info("Rates file is empty")
        rates_df["raw"] = None
        rates_df["score"] = None
        rates_df.to_csv(output, sep="\t", index=False)
        sys.exit(0)

    # Depending on the gff, chromosome can be defined as "chrN" or just "N"
    if str(rates_df.iloc[0].chrom).startswith("chr"):
        add_chr = True
    else:
        add_chr = False

    return rates_df, add_chr


def remove_duplicates(block_dbnsfp):
    """
    Looking at dbNSFP it looks like we can have two records for a same variant if it leads to
    a different amino acid.
    For now we prioritise the record that matches the transcript in the input GFF (if any) and,
    if several, the record with less missing annotations.


    Args:
        block_dbnsfp (pd.DataFrame): subset of the annotated data frame
    """

    indices_to_remove = list()
    for pos, min_df in block_dbnsfp.loc[block_dbnsfp[["chrom", "pos", "ref", "alt"]].duplicated(keep=False)].groupby(
        ["pos", "alt"]
    ):

        indices = min_df.index
        if min_df["transcript_in_dbnsfp"].sum() != 0:
            min_df = min_df.loc[min_df["transcript_in_dbnsfp"] == True].copy()

        # Keep the record with the most values annotated
        min_df["nb_missing_values"] = (min_df != ".").sum(axis=1)
        keep_idx = min_df.sort_values("nb_missing_values", ascending=False).index[0]

        indices_to_remove = indices_to_remove + list(set(indices) - set([keep_idx]))

    block_dbnsfp = block_dbnsfp.drop(indices_to_remove)

    return block_dbnsfp


def extract_ensembl_gene_id_without_version(gff_db):
    """
    It might happen that the user input file contains ENSEMBL gene ids without version number
    when the GFF file does contain them. In that case we need to map the two together.

    Args:
        gff_db (gffutils.FeatureDB): gffutils database

    Returns:
        dict: maps ENSG with version to ENSG without version (e.g. {ENSG00000010404 : ENSG00000010404.1})
    """

    ensembl_gene_id_map_version = dict()
    for gene in gff_db.features_of_type("gene", order_by="start"):
        ensembl_gene_id_map_version[gene.id.split(".")[0]] = gene.id

    return ensembl_gene_id_map_version


@click.command()
@click.argument("rates_dnm")
@click.argument("dbnsfp")
@click.argument("output")
@click.option("-c", "--columns", default="")
@click.option("-C", "--columns-file", default="")
@click.option("--gff", default="")
def annotate_dbnsfp(rates_dnm, dbnsfp, output, columns, columns_file, gff):
    """
    Annotatate a rates/DNM file with CEP scores from dbNSFP.
    When multiple scores exist for a variant (e.g. multiple transcripts, overlapping genes),
    the maximum score among transcripts found in the GFF are retrieved.
    When no GFF are provided, the maximum score for the gene associated to the variant is retrieved.

    Args:
        rates_dnm (str): Variants file (e.g. rates, DNM) starting with the following columns [gene_id,chrom,pos,ref,alt]
        dbnsfp (str): dbNSFP genome wide file
        output (str): Output file (annotated dataframe)
        columns(str) : Annotations to extract form dbNSFP (comma separated)
        columns_file (str): File listing which annotations to extract from dbNSFP
        gff(str) : Gff file or gffutils database
    """

    init_log()
    logger = logging.getLogger("logger")

    # Load rates/DNM file
    df, add_chr = load_rates_file(rates_dnm, output)
    rates_df_columns = list(df.columns)

    # Load dbnsfp file
    dbnsfp_df = pysam.TabixFile(dbnsfp, encoding="utf-8")

    # Load gff file/DB (to match annotation based on user selected transcripts)
    if gff:
        gff_db = load_gff(gff)
        ensembl_gene_id_map_version = extract_ensembl_gene_id_without_version(gff_db)
    else:
        gff_db = None

    # Retrieve columns to extract from dbnsfp_df
    dbnsfp_columns_indices, dbnsfp_columns_names = extract_columns_from_dbNFP(dbnsfp_df, columns, columns_file)

    # If some columns are already found in the input file, we prefix the new ones with "dbnsfp"
    existing_columns = check_columns(rates_df_columns, dbnsfp_columns_names)
    dbnsfp_columns_names = [f"{col}_dbnsfp" if col in existing_columns else col for col in dbnsfp_columns_names]

    # For each gene
    list_merged_df = list()
    for gene_id, gene_rates_df in df.groupby("gene_id"):

        logger.info(f"Annotating {gene_id}")

        chrom = str(gene_rates_df.chrom.values[0]).replace("chr", "")

        # Split each gene in contiguous block (i.e. exons) and load dbNSFP scores
        list_block_df = list()
        for _, block in groupby(sorted(set(gene_rates_df["pos"])), key=lambda n, c=count(): n - next(c)):
            start, end = as_range(list(block))
            try:
                block_dbnsfp = load_dbnsfp(
                    dbnsfp_df,
                    chrom,
                    start - 1,
                    end,
                    dbnsfp_columns_indices,
                    gene_id,
                    gff_db,
                    ensembl_gene_id_map_version,
                )
                if not block_dbnsfp.empty:
                    block_dbnsfp = remove_duplicates(block_dbnsfp)
                    list_block_df.append(block_dbnsfp)
            except ValueError:
                continue

        # Merge annotations from several blocks together
        if list_block_df:
            gene_dbnsfp_df = pd.concat(list_block_df)
            if add_chr:
                gene_dbnsfp_df.chrom = "chr" + gene_dbnsfp_df.chrom

            merged_gene_df = gene_rates_df.merge(gene_dbnsfp_df, how="left", on=["chrom", "pos", "ref", "alt"])
        else:
            merged_gene_df = gene_rates_df

        list_merged_df.append(merged_gene_df)

    # Combine results for all genes
    merged_df = pd.concat(list_merged_df)
    merged_df.columns = rates_df_columns + ["transcript_in_dbnsfp"] + dbnsfp_columns_names
    merged_df.to_csv(output, sep="\t", index=False)


if __name__ == "__main__":
    merged_df = annotate_dbnsfp()



#!/usr/bin/env python

import pandas as pd
import click
from pysam import VariantFile
from itertools import groupby, count
import numpy as np


def load_gnomad(gnomad, chrom, start, end):
    """
    Access to specific region in gnomad VCF file and builds a data frame
    Args:
        gnomad (pysam.TabixFile): CADD score file
        chrom (str): chromosome identifier
        start (int): beginning position of the region
        end (int): terminating position of the region
    """
    gnomad_data = []
    for record in gnomad.fetch(chrom, start, end):
        for i, alt in enumerate(record.alts):
            if "AF" in record.info and record.info["AF"][i] is not None:
                dic = {
                    "chrom": str(record.chrom),
                    "pos": int(record.pos),
                    "ref": str(record.ref),
                    "alt": str(alt),
                    "maf": float(record.info["AF"][i]),
                }
                gnomad_data.append(dic)
    return pd.DataFrame(gnomad_data)


def as_range(region):
    """
    Returns genomic region boundaries

    Args:
        region (list): genomic region positions

    Returns:
        tuple: beginning and terminating position of the genomic region
    """
    l = list(region)
    return l[0], l[-1]


@click.command()
@click.argument("rates")
@click.argument("gnomad")
@click.argument("output")
def annotate_gnomad(rates, gnomad, output):
    """Adds gnomad MAF to rates file

    Args:
        rates (str): Path to rates file
        gnomad (str): Path to gnomad file
        output (str): Path to output file (merged dataframe)
    """

    # Load rates file
    rates_df = pd.read_table(rates)

    # Load gnomad VCF
    gnomad_df = VariantFile(gnomad)

    list_merged_df = list()
    for gene_id, gene_rates_df in rates_df.groupby("gene_id"):
        chrom = gene_rates_df.chrom.values[0]

        # Split each gene in contiguous block (i.e. exons) and load gnomad data
        # (memory and performance issue)
        list_block_df = list()
        for _, block in groupby(sorted(set(gene_rates_df["pos"])), key=lambda n, c=count(): n - next(c)):
            start, end = as_range(block)

            block_gnomad_df = load_gnomad(gnomad_df, chrom, start - 1, end)
            list_block_df.append(block_gnomad_df)

        gene_gnomad_df = pd.concat(list_block_df)

        # Merge rates with Gnomad data
        if gene_gnomad_df.empty:
            merged_gene_df = gene_rates_df
            gene_rates_df["maf"] = np.nan
        else:
            merged_gene_df = gene_rates_df.merge(gene_gnomad_df, how="left", on=["chrom", "pos", "ref", "alt"])
        list_merged_df.append(merged_gene_df)

    # Combine results for all genes
    merged_df = pd.concat(list_merged_df)
    merged_df.to_csv(output, sep="\t", index=False)


if __name__ == "__main__":
    annotate_gnomad()



import numpy as np
import click

from scipy import stats
from joblib import Parallel, delayed
from denovowest.utils.params import DEFAULT_MAX_NB_MUTATIONS_SIM, DEFAULT_MIN_NB_SIM


def calc_p0(mu, obs_sum_scores):
    """
    Exact calculate P(S >= obs_sum_scores | N = 0)P(N = 0)

    Args:
        mu (float): poisson parameter correpsponding to the sum of the mutation rates
        obs_sum_scores (float): sum of observed DNM scores
    """

    if obs_sum_scores == 0:
        p0 = 1 * stats.poisson.pmf(0, mu)
    else:
        p0 = 0
    return p0


def calc_p1(mu, obs_sum_scores, rates, score_column):
    """
    Exact calculate P(S >= obs_sum_scores | N = 1)P(N = 1)

    Args:
        mu (float): poisson parameter correpsponding to the sum of the mutation rates
        obs_sum_scores (float): sum of observed DNM scores
        rates (pd.DataFrame): all possible mutations annotated
        score_column (str) : CEP scores

    """

    # Get the proportion of putative variants that have a score greater than the observed sum of scores
    p1c = rates["prob"][rates[score_column] >= obs_sum_scores].sum() / rates["prob"].sum()

    # Weight it by the probability of observing one mutation given the poisson rate (expected number of mutations)
    p1 = p1c * stats.poisson.pmf(1, mu)

    return p1


def calc_pn(mu, obs_sum_scores, rates, nb_mutation_poisson, nsim, scores_sorted, score_column):
    """
    Simulation to approximate  P(S >= s_obs | N = n)P(N = n)

    Args:
        mu (float): poisson parameter correpsponding to the sum of the mutation rates
        obs_sum_scores (float): sum of observed DNM scores
        rates (pd.DataFrame): all possible mutations annotated
        nb_mutation_poisson (int): number of mutations to draw
        nsim (int) : number of simulations to perform
        scores_sorted (list) : rates file variants sorted by score
        score_column (str) : CEP scores


    Returns :
        Tuple : (pn, nsim), pn represents the probability of observing a score greater than or equal to the observed score if we select randomly nb_mutation_poisson mutations
        nsim is the number of simulations performed
    """

    # Compute the probability of observing nb_mutation_poisson mutation given the poisson rate (expected number of mutations)
    pndnm = stats.poisson.pmf(nb_mutation_poisson, mu)

    # Scaling the number of simulations to be performed based on the probability of observing nb_mutation_poisson DNMs
    # TODO : see if there is a room for improvement here too
    nsim = max([int(round(nsim * pndnm)), DEFAULT_MIN_NB_SIM])

    # If pndm (see above) is really low there is no point in calculating a combined p-value as it will be epsilon
    if pndnm < 10 ** (-12):
        pscore = 0
        nsim = 0
    # If the top nb_mutation_poisson scores are not enough to reach the observed sum of scores, then there is no possible nb_mutation_poisson combination that
    # will achieve a higher score and the p-value is 0
    elif np.sum(scores_sorted[-nb_mutation_poisson:]) < obs_sum_scores:
        pscore = 0.0
        nsim = 0

    # On the contrary, if the lowest nb_mutation_poisson scores are enough to reach the observed sum of scores, then there is no possible nb_mutation_poisson combination that
    # will achieve a lower score and the p-value is 1
    elif np.sum(scores_sorted[0:nb_mutation_poisson] >= obs_sum_scores):
        pscore = 1.0
        nsim = 0

    # Otherwise, we simulate the cumulated scores for nb_mutation_poisson randomly picked mutations nsim times and calculate the proportion of simulations
    # for which we obtain a score greater than or equal to the observed score
    else:
        s = sim_score(mu, obs_sum_scores, rates, nb_mutation_poisson, nsim, score_column)
        pscore = float(s) / nsim

    # This probability is adjusted by the probability of actually observing nb_mutation_poisson mutations given the poisson rate (expected number of mutations)
    pn = pndnm * pscore

    return (pn, nsim)


def sim_score(mu, obs_sum_scores, rates, nb_mutation_poisson, nsim, score_column):
    """
    Draws nsim times nb_mutation_poisson mutations from set of all possible mutations according to their mutation rate.
    Count how many times their cumulated scores is greater than or equal to the observed sum of scores.

    Args:
        mu (float): poisson parameter correpsponding to the sum of the mutation rates
        obs_sum_scores (float): sum of observed DNM scores
        rates (pd.DataFrame): all possible mutations annotated
        nb_mutation_poisson (int): number of mutations to draw
        nsim (int) : number of simulations to perform
        score_column(str) : column to extract the scores from
    """

    # Precompute probabilities
    probabilities = rates["prob"] / mu
    scores = rates[score_column].values

    def simulate_chunk(chunk_size):
        """Simulate one chunk of scores."""
        simulated_scores = np.sum(np.random.choice(scores, (nb_mutation_poisson, chunk_size), p=probabilities), axis=0)
        return np.sum(simulated_scores >= obs_sum_scores)

    # Split simulations into chunks
    split_sim = 100000
    num_full_chunks = nsim // split_sim
    remaining_simulations = nsim % split_sim

    # Run full chunks in parallel
    ctx = click.get_current_context()
    full_chunk_results = Parallel(n_jobs=ctx.params["jobs"])(
        delayed(simulate_chunk)(split_sim) for _ in range(num_full_chunks)
    )

    # Run remaining simulations (if any)
    remaining_result = simulate_chunk(remaining_simulations) if remaining_simulations > 0 else 0

    # Sum up results
    nb_more_extreme_scores = sum(full_chunk_results) + remaining_result

    return nb_more_extreme_scores


def get_pvalue(rates, obs_sum_scores, nsim, pvalcap, nb_observed_mutations, score_column):
    """
    Calculate the p-value from the enrichment simulation test

    Args:
        rates (DataFrame): all possible SNVs annotated
        obs_sum_scores (float): sum of observed DNM scores
        nsim (int): Number of simulations to perform.
        pvalcap (float): P-value threshold to stop simulations.
        nb_observed_mutations (int) : Number of observed mutations in the current gene
        score_column (str) : CEP scores


    Returns:
        tuple: A tuple containing the p-value, simulation information, and expected score.
    """

    # The poisson rate is the sum of the adjusted mutation rates of all possible mutations in the gene
    mu = rates["prob"].sum()

    # The expected score is the sum of all possible mutations scores weighted by their mutation rates (already adjusted for cohort size)
    exp_sum_scores = np.sum(rates["prob"] * rates[score_column])

    # If the observed score is already lower than the expected one, no need to run the simulation
    if obs_sum_scores < exp_sum_scores:
        ptot = 1
        info = "0|0|True|observed < expected, pvalue set at 1"
        return ptot, info, exp_sum_scores

    # We sort the scores in order to use stopping rules that improve the speed of the simulations
    scores_sorted = np.sort(rates[score_column])

    # Calculate the probability of seeing a similar or more extreme observed score when selecting 0 or 1 mutation
    p0 = calc_p0(mu, obs_sum_scores)
    p1 = calc_p1(mu, obs_sum_scores, rates, score_column)

    # If the gene has a really high number of observed mutations, we increase the default threshold to two times the number of observed mutations
    nb_putative_mutations_sim = max(2 * nb_observed_mutations, DEFAULT_MAX_NB_MUTATIONS_SIM)

    # For long genes with a high number of expected mutations, it is a better strategy
    # to start with a poisson mutation rate close to the expected mutation rate,
    # that way, as most genes will not have an enrichment, we can discard them more quickly
    if mu > 50:
        range_mutation = list(diverging_range(int(mu), 2, nb_putative_mutations_sim))
        sequential = False
    else:
        range_mutation = range(2, nb_putative_mutations_sim)
        sequential = True

    # Running simulations
    ptot = p0 + p1
    nbsim_tot = 0
    info = ""
    for nb_mutation_poisson in range_mutation:

        # Calculate the probability of seeing a similar or more extreme observed gene score | nb_mutation_poisson mutations
        pi, nb_sim = calc_pn(mu, obs_sum_scores, rates, nb_mutation_poisson, nsim, scores_sorted, score_column)
        ptot = ptot + pi
        nbsim_tot = nbsim_tot + nb_sim

        # If we are increasing the number of mutation in sequential order we can
        # stop if the poisson p-value becomes too low from a certain point
        if sequential:
            # Probability of observing more than nb_mutation_poisson event given mu
            picdf = 1 - stats.poisson.cdf(nb_mutation_poisson, mu)

            # If this probability is too low, the resulting p-values will
            if (picdf < 10 ** (-12)) and (nb_mutation_poisson > mu):
                info = "probability of observing >= " + str(nb_mutation_poisson) + " mutations is too small"
                break

        # if p value is over threshold then stop going further
        if ptot > pvalcap:
            info = "pvalue > " + str(pvalcap) + ", stop simulations"
            break

    # Set min p-value to 10^-14
    if ptot == 0:
        ptot = 10 ** (-14)
        info = "p value was 0, set at 10^-14"

    # Add the number of simulations performed to the simulation information
    info = f"{nbsim_tot}|{nb_mutation_poisson}|{sequential}|{info}"

    return (ptot, info, exp_sum_scores)


def diverging_range(median, min_val, max_val):
    """
    Yields a sequence of integers starting from `median` and alternating outward
    in both directions (e.g., median+1, median-1, median+2, median-2, ...),
    within the bounds [min_val, max_val].

    Example:
        list(diverging_range(5, 2, 8))
        → [5, 6, 4, 7, 3, 8, 2]
    """

    yield median  # Start by yielding the median value
    offset = 1  # Initialize the offset from the median

    while True:
        lower = median - offset  # Step downward
        upper = median + offset  # Step upward
        did_yield = False  # Track if any value was yielded this round

        # If upper value is within bounds, yield it
        if upper <= max_val:
            yield upper
            did_yield = True

        # If lower value is within bounds, yield it
        if lower >= min_val:
            yield lower
            did_yield = True

        # Stop the loop if no values were yielded in this iteration
        if not did_yield:
            break

        # Increase the distance from the median for the next iteration
        offset += 1



#############################
# SHARED BY SEVERAL MODULES #
#############################

# Number of adjacent nucleotides to consider when retrieving CDS
CDS_OFFSET = 50


# Bcftools consequences ordered by severity
CONSEQUENCES_SEVERITIES = {
    "intergenic": 1,
    "feature_truncation": 2,
    "feature_elongation": 2,
    "regulatory": 3,
    "TF_binding_site": 4,
    "TFBS": 4,
    "downstream": 5,
    "upstream": 5,
    "non_coding_transcript": 6,
    "non_coding": 6,
    "intron": 7,
    "NMD_transcript": 7,
    "non_coding_transcript_exon": 8,
    "5_prime_utr": 9,
    "3_prime_utr": 9,
    "coding_sequence": 10,
    "mature_miRNA": 10,
    "stop_retained": 11,
    "start_retained": 11,
    "synonymous": 11,
    "incomplete_terminal_codon": 12,
    "splice_region": 13,
    "missense": 14,
    "inframe": 14,
    "inframe_insertion": 14,
    "inframe_deletion": 14,
    "protein_altering": 14,
    "transcript_amplification": 15,
    "exon_loss": 16,
    "disruptive": 17,
    "start_lost": 18,
    "stop_lost": 18,
    "stop_gained": 18,
    "frameshift": 18,
    "splice_acceptor": 19,
    "splice_donor": 19,
    "transcript_ablation": 20,
}


##############
# RATES #
##############


# Per generation mutation rate scaling factors taken from https://github.com/vseplyarskiy/Roulette/tree/main/adding_mutation_rate
# Divided by 2 because the provided rate are per diploid genomes
ROULETTE_SCALING_FACTOR = 1.015e-7 / 2
CARLSON_SCALING_FACTOR = 2.086e-9 / 2


##############
# SIMULATION #
##############

# We use meta category in the simulation (e.g. start lost are assimilated to missense)
# Part of the mapping is legacy from DNW v1 where consequences were using mixing two format
CONSEQUENCES_MAPPING = {
    "frameshift_variant": "frameshift",
    "frameshift": "frameshift",
    "inframe_insertion": "inframe",
    "inframe_deletion": "inframe",
    "missense_variant": "missense",
    "missense": "missense",
    "stop_gained": "nonsense",
    "synonymous_variant": "synonymous",
    "splice_acceptor_variant": "splice_lof",
    "splice_donor_variant": "splice_lof",
    "splice_acceptor": "splice_lof",
    "splice_donor": "splice_lof",
    "splice_region_variant": "splice_region",
    "splice_region": "splice_region",
    "conserved_exon_terminus_variant": "splice_lof",
    "start_lost": "missense",
    "stop_lost": "missense",
    "stop_retained": "synonymous",
    "synonymous": "synonymous",
    "nonsense": "nonsense",
    "splice_lof": "splice_lof",
    "inframe": "inframe",
}

# Default maximum number of expected mutation to test for using the simulation approach
DEFAULT_MAX_NB_MUTATIONS_SIM = 250

# Minimum number of simulation per batch
DEFAULT_MIN_NB_SIM = 10**7

# These inframe/missense and frameshift/nonsense ratios were observed on several databases
# such ad EXAC, gnomAD, TOPMED or DDD, and are used as proxys to estimate inframe/frameshift mutation rates
INFRAME_MISSENSE_RATIO = 0.03
FRAMESHIFT_NONSENSE_RATIO = 1.3
